# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from collections import defaultdict
from functools import singledispatchmethod

import numpy as np
from compiler_config.config import CompilerConfig, ErrorMitigationConfig, ResultsFormatting

from qat.core.config.configure import get_config
from qat.core.pass_base import ValidationPass
from qat.core.result_base import ResultManager
from qat.ir.instruction_builder import InstructionBuilder, QuantumInstructionBuilder
from qat.ir.instructions import FrequencySet, FrequencyShift, Instruction, Repeat, Return
from qat.ir.measure import Acquire, AcquireMode, PostProcessing, ProcessAxis
from qat.ir.waveforms import Pulse
from qat.middleend.passes.analysis import ActivePulseChannelResults
from qat.model.device import PhysicalChannel
from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.target_data import TargetData
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class InstructionValidation(ValidationPass):
    """
    Validates instructions against the hardware, including whether:
    - they fit in memory,
    - pulse durations are within the allowed limits and
    - acquire instructions are only performed on acquire channels..
    """

    def __init__(
        self,
        target_data: TargetData,
        model: PhysicalHardwareModel,
        pulse_duration_limits: bool | None = None,
        *args,
        **kwargs,
    ):
        """
        :param target_data: Target-related information.
        :param pulse_duration_limits: Whether to check the pulse duration limits.
            If None, uses the default from the QatConfig.
        :param instruction_memory_size: The maximum number of instructions that can
            be run in a single shot. If None, uses the default from QatConfig.
        """
        self.instruction_memory_size = target_data.instruction_memory_size
        self.pulse_duration_max = max(
            target_data.QUBIT_DATA.pulse_duration_max,
            target_data.RESONATOR_DATA.pulse_duration_max,
        )
        self.pulse_duration_min = min(
            target_data.QUBIT_DATA.pulse_duration_min,
            target_data.RESONATOR_DATA.pulse_duration_min,
        )
        self.pulse_duration_limits = (
            pulse_duration_limits
            if pulse_duration_limits is not None
            else get_config().INSTRUCTION_VALIDATION.PULSE_DURATION_LIMITS
        )
        self._channel_data = self._get_channel_data(model)

    @staticmethod
    def _get_channel_data(model):
        data = {"acquire": [], "non-acquire": {}}
        for q_id, qubit in model.qubits.items():
            data["acquire"].append(qubit.acquire_pulse_channel.uuid)
            for pulse_channel in qubit.all_pulse_channels + [qubit.measure_pulse_channel]:
                data["non-acquire"][pulse_channel.uuid] = [pulse_channel.pulse_type, q_id]
        return data

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        instruction_length = ir.number_of_instructions
        if instruction_length > self.instruction_memory_size:
            raise ValueError(
                f"Program with {instruction_length} instructions too large to be run "
                f"in a single block on current hardware."
            )

        for inst in ir.instructions:
            self._validate_instruction(inst)

        return ir

    @singledispatchmethod
    def _validate_instruction(self, instruction: Instruction):
        pass

    @_validate_instruction.register(Pulse)
    def _(self, instruction: Pulse):
        duration = instruction.duration
        if duration > self.pulse_duration_max or duration < self.pulse_duration_min:
            if self.pulse_duration_limits:
                # Do not throw error if we specifically disabled the limit checks.
                raise ValueError(
                    f"Waveform width must be between {self.pulse_duration_min} s "
                    f"and {self.pulse_duration_max} s. "
                    f"Given has a width: {instruction.duration} s"
                )

    @_validate_instruction.register(Acquire)
    def _(self, instruction: Acquire):
        if instruction.target not in self._channel_data["acquire"]:
            channel = self._channel_data["non-acquire"][instruction.target]
            raise ValueError(
                f"Cannot perform an acquire on the {channel[0]} pulse channel for qubit {channel[1]}"
            )


class DynamicFrequencyValidation(ValidationPass):
    """Validates the setting or shifting frequencies does not move the intermediate
    frequency of a pulse channel outside the allowed limits."""

    def __init__(self, model: PhysicalHardwareModel, target_data: TargetData):
        """Instantiate the pass with a hardware model.

        :param model: The hardware model.
        :param target_data: Target-related information.
        """
        self.model = model
        self._is_resonator = self._create_resonator_map(model)

        # Extract limits from the target data.
        self.qubit_if_freq_limits = (
            target_data.QUBIT_DATA.pulse_channel_if_freq_min,
            target_data.QUBIT_DATA.pulse_channel_if_freq_max,
        )
        self.resonator_if_freq_limits = (
            target_data.RESONATOR_DATA.pulse_channel_if_freq_min,
            target_data.RESONATOR_DATA.pulse_channel_if_freq_max,
        )

    @staticmethod
    def _create_resonator_map(model: PhysicalHardwareModel) -> dict[str, bool]:
        """Creates a map to lookup if physical channels belong to resonators.."""
        res_map = {}
        for qubit in model.qubits.values():
            res_map[qubit.physical_channel.uuid] = False
            res_map[qubit.resonator.physical_channel.uuid] = True
        return res_map

    def _validate_frequency_shifts(self, physical_channel_id: str, ifs: list[float]):
        """Validates that the frequency shifts do not exceed the allowed limits."""

        if_limits = (
            self.qubit_if_freq_limits
            if not self._is_resonator[physical_channel_id]
            else self.resonator_if_freq_limits
        )

        return [
            if_value
            for if_value in ifs
            if not (if_limits[0] <= np.abs(if_value) <= if_limits[1])
        ]

    @singledispatchmethod
    def _calculate_frequency(
        self,
        instruction: Instruction,
        ir: QuantumInstructionBuilder,
        ifs: dict[str, list[float]],
        physical_channel_ids: dict[str, str],
    ):
        pass

    @_calculate_frequency.register(FrequencyShift)
    def _(
        self,
        instruction: FrequencyShift,
        ir: QuantumInstructionBuilder,
        ifs: dict[str, list[float]],
        physical_channel_ids: dict[str, str],
    ):
        frequencies = ifs[instruction.target]
        pulse_channel = ir.get_pulse_channel(instruction.target)
        physical_channel = self.model.physical_channel_with_id(
            pulse_channel.physical_channel_id
        )
        freq = (
            frequencies[-1]
            if len(frequencies) > 0
            else pulse_channel.frequency - physical_channel.baseband.frequency
        )
        frequencies.append(freq + instruction.frequency)
        physical_channel_ids[instruction.target] = physical_channel.uuid

    @_calculate_frequency.register(FrequencySet)
    def _(
        self,
        instruction: FrequencySet,
        ir: QuantumInstructionBuilder,
        ifs: dict[str, list[float]],
        physical_channel_ids: dict[str, str],
    ):
        physical_channel = self.model.physical_channel_with_id(
            ir.get_pulse_channel(instruction.target).physical_channel_id
        )
        ifs[instruction.target].append(
            instruction.frequency - physical_channel.baseband.frequency
        )
        physical_channel_ids[instruction.target] = physical_channel.uuid

    def run(
        self, ir: QuantumInstructionBuilder, *args, **kwargs
    ) -> QuantumInstructionBuilder:
        """:param ir: The list of instructions stored in an :class:`QuantumInstructionBuilder`."""

        ifs = defaultdict(list)
        physical_channel_ids = defaultdict(str)
        for instruction in ir.instructions:
            self._calculate_frequency(instruction, ir, ifs, physical_channel_ids)

        violations = []
        for pulse_channel_id, if_values in ifs.items():
            physical_channel_id = physical_channel_ids[pulse_channel_id]
            if if_violations := self._validate_frequency_shifts(
                physical_channel_id, if_values
            ):
                qubit = self.model.qubit_for_physical_channel_id(
                    physical_channel_ids[pulse_channel_id]
                )
                device_name = (
                    "Resonator " if self._is_resonator[physical_channel_id] else "Qubit "
                )
                device_name += str(self.model.index_of_qubit(qubit))

                if (
                    hw_pc := self.model.pulse_channel_with_id(pulse_channel_id)
                ) is not None:
                    pc_name = f"{hw_pc.pulse_type}"
                else:
                    pc_name = f"Custom({pulse_channel_id})"

                violations.append(
                    f"The IF of {device_name} {pc_name} pulse channel is set or "
                    f"shifted to values {if_violations} that exceed the allowed limits."
                )

        if len(violations) > 0:
            raise ValueError(
                "Dynamic frequency validation failed with the following violations:\n"
                + "\n".join(violations)
            )
        return ir


class NoMidCircuitMeasurementValidation(ValidationPass):
    """
    Validates that there are no mid-circuit measurements by checking that no qubit
    has an acquire instruction that is later followed by a pulse instruction.
    """

    def __init__(
        self,
        model: PhysicalHardwareModel,
        no_mid_circuit_measurement: bool | None = None,
        *args,
        **kwargs,
    ):
        """
        :param model: The hardware model.
        :param no_mid_circuit_measurement: Whether mid-circuit measurements are allowed.
            If None, uses the default from the QatConfig.
        """
        self.model = model
        self.no_mid_circuit_measurement = (
            no_mid_circuit_measurement
            if no_mid_circuit_measurement is not None
            else get_config().INSTRUCTION_VALIDATION.NO_MID_CIRCUIT_MEASUREMENT
        )

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """
        :param ir: The intermediate representation (IR) :class:`InstructionBuilder`.
        """
        consumed_acquire_pc: set[str] = set()

        if not self.no_mid_circuit_measurement:
            return ir

        drive_acq_pc_map = {
            qubit.drive_pulse_channel.uuid: qubit.acquire_pulse_channel.uuid
            for qubit in self.model.qubits.values()
        }

        for instr in ir:
            if isinstance(instr, Acquire):
                consumed_acquire_pc.add(instr.target)

            # Check if we have a measure in the middle of the circuit somewhere.
            elif isinstance(instr, Pulse):
                acq_pc = drive_acq_pc_map.get(instr.target, None)

                if acq_pc and acq_pc in consumed_acquire_pc:
                    raise ValueError(
                        "Mid-circuit measurements currently unable to be used."
                    )
        return ir


class ReadoutValidation(ValidationPass):
    """Validates that the post-processing instructions do not have an invalid sequence.

    Extracted from :meth:`qat.purr.backends.live.LiveDeviceEngine.validate`.
    """

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""
        acquire_modes = {}
        for inst in ir:
            if isinstance(inst, Acquire):
                acquire_modes[inst.output_variable] = inst.mode
            if isinstance(inst, PostProcessing):
                acquire_mode = acquire_modes.get(inst.output_variable, None)
                self._post_processing_options_handling(inst, acquire_mode)
        return ir

    @staticmethod
    def _post_processing_options_handling(
        inst: PostProcessing, acquire_mode: AcquireMode | None
    ):
        if acquire_mode == AcquireMode.SCOPE and ProcessAxis.SEQUENCE in inst.axes:
            raise ValueError(
                "Invalid post-processing! Post-processing over SEQUENCE is "
                "not possible after the result is returned from hardware "
                "in SCOPE mode!"
            )
        elif acquire_mode == AcquireMode.INTEGRATOR and ProcessAxis.TIME in inst.axes:
            raise ValueError(
                "Invalid post-processing! Post-processing over TIME is not "
                "possible after the result is returned from hardware in "
                "INTEGRATOR mode!"
            )
        elif acquire_mode == AcquireMode.RAW:
            raise ValueError(
                "Invalid acquire mode! The live hardware doesn't support RAW acquire mode!"
            )
        elif acquire_mode is None:
            raise ValueError(
                f"No AcquireMode found with output variable {inst.output_variable},"
                f"ensure PostProcessing output_variable matches an Acquire output_variable with a"
                f"valid AcquireMode selected."
            )


class HardwareConfigValidity(ValidationPass):
    """Validates the :class:`CompilerConfig` against the hardware model."""

    def __init__(self, hardware_model: PhysicalHardwareModel, max_shots: int | None = None):
        """Instantiate the pass with a hardware model.

        :param hardware_model: The hardware model.
        :param max_shots: The maximum number of shots allowed for a single task.
                If None, uses the default from the QatConfig.
        """
        self.hardware_model = hardware_model
        self.max_shots = (
            max_shots if max_shots is not None else get_config().MAX_REPEATS_LIMIT
        )

    def run(
        self,
        ir: InstructionBuilder,
        *args,
        compiler_config: CompilerConfig,
        **kwargs,
    ):
        self._validate_shots(compiler_config)
        self._validate_error_mitigation(self.hardware_model, compiler_config)
        return ir

    def _validate_shots(self, compiler_config: CompilerConfig):
        shots = compiler_config.repeats
        if shots is not None and shots > self.max_shots:
            raise ValueError(
                f"Number of shots in compiler config {compiler_config.repeats} exceeds max "
                f"number of shots {self.max_shots}."
            )

    def _validate_error_mitigation(
        self, hardware_model: PhysicalHardwareModel, compiler_config: CompilerConfig
    ):
        if (
            compiler_config.error_mitigation
            and compiler_config.error_mitigation != ErrorMitigationConfig.Empty
        ):
            if not hardware_model.error_mitigation.is_enabled:
                raise ValueError("Error mitigation not calibrated on this hardware model.")

            if ResultsFormatting.BinaryCount not in compiler_config.results_format:
                raise ValueError(
                    "BinaryCount format required for readout error mitigation."
                )


class FrequencySetupValidation(ValidationPass):
    """Validates the baseband frequencies and intermediate frequencies of pulse channels
    against the target data.
    """

    def __init__(self, model: PhysicalHardwareModel, target_data: TargetData):
        """Instantiate the pass with a hardware model.

        :param model: The hardware model.
        :param target_data: Target-related information.
        """

        # Extract limits from the target data.
        self.qubit_lo_freq_limits = (
            target_data.QUBIT_DATA.pulse_channel_lo_freq_min,
            target_data.QUBIT_DATA.pulse_channel_lo_freq_max,
        )
        self.resonator_lo_freq_limits = (
            target_data.RESONATOR_DATA.pulse_channel_lo_freq_min,
            target_data.RESONATOR_DATA.pulse_channel_lo_freq_max,
        )
        self.qubit_if_freq_limits = (
            target_data.QUBIT_DATA.pulse_channel_if_freq_min,
            target_data.QUBIT_DATA.pulse_channel_if_freq_max,
        )
        self.resonator_if_freq_limits = (
            target_data.RESONATOR_DATA.pulse_channel_if_freq_min,
            target_data.RESONATOR_DATA.pulse_channel_if_freq_max,
        )
        self._physical_channel_data = self._get_physical_channel_data(model)
        self._model = model

    def _get_physical_channel_data(self, model: PhysicalHardwareModel) -> dict[str, dict]:
        """Extracts physical channel data from the model."""
        physical_channel_data = {}
        for q_index, qubit in model.qubits.items():
            physical_channel = qubit.physical_channel
            physical_channel_data[physical_channel.uuid] = (
                self._populate_physical_channel_data(physical_channel, False, q_index)
            )

            physical_channel = qubit.resonator.physical_channel
            physical_channel_data[physical_channel.uuid] = (
                self._populate_physical_channel_data(physical_channel, True, q_index)
            )

        return physical_channel_data

    def _populate_physical_channel_data(
        self, physical_channel: PhysicalChannel, is_resonator: bool, qubit_index: int
    ) -> dict:
        physical_channel_data = {
            "baseband_frequency": physical_channel.baseband.frequency,
            "is_resonator": is_resonator,
            "is_valid": self._validate_baseband_frequency(
                physical_channel.baseband.frequency, is_resonator
            ),
            "index": qubit_index,
        }
        return physical_channel_data

    def _validate_baseband_frequency(self, frequency, is_resonator: bool) -> bool:
        """Validates that the frequencies in the dictionary are within the specified range."""
        lower, upper = self._get_lo_freq_range(is_resonator)
        return bool(lower <= frequency <= upper)

    def _validate_pulse_channel_if(self, frequency: float, is_resonator: bool) -> bool:
        """Validates that the pulse channel frequencies are within the specified range."""
        lower, upper = self._get_if_freq_range(is_resonator)
        return bool(lower <= np.abs(frequency) <= upper)

    def _get_lo_freq_range(self, is_resonator: bool) -> tuple[float, float]:
        """Returns the frequency range for the given type of channel."""
        return (
            self.qubit_lo_freq_limits if not is_resonator else self.resonator_lo_freq_limits
        )

    def _get_if_freq_range(self, is_resonator: bool) -> tuple[float, float]:
        """Returns the IF frequency range for the given type of channel."""
        return (
            self.qubit_if_freq_limits if not is_resonator else self.resonator_if_freq_limits
        )

    def _find_physical_channel_violations(self, physical_channels) -> list[str] | None:
        """Finds violations in the physical channels based on their baseband frequencies."""
        violations = []
        for physical_channel in physical_channels:
            data = self._physical_channel_data.get(physical_channel, None)
            if not data["is_valid"]:
                freq_range = self._get_lo_freq_range(data["is_resonator"])
                baseband_freq = data["baseband_frequency"]
                name = "Resonator" if data["is_resonator"] else "Qubit"
                name += f" {data['index']}"
                violations.append(
                    f"Physical channel for {name} with baseband frequency "
                    f"{baseband_freq} is out of the valid "
                    f"range {freq_range}."
                )
        return violations

    def _find_pulse_channel_violations(
        self, pulse_channels: set[str], ir: QuantumInstructionBuilder
    ) -> list[str] | None:
        """Finds violations in the pulse channels based on their IF frequencies."""
        violations = []
        for pulse_channel_id in pulse_channels:
            pulse_channel = ir.get_pulse_channel(pulse_channel_id)
            physical_channel_id = pulse_channel.physical_channel_id
            physical_channel_data = self._physical_channel_data[physical_channel_id]

            if_freq = pulse_channel.frequency - physical_channel_data["baseband_frequency"]
            valid = self._validate_pulse_channel_if(
                if_freq, physical_channel_data["is_resonator"]
            )
            if not valid:
                name = "Resonator" if physical_channel_data["is_resonator"] else "Qubit"
                name += f" {physical_channel_data['index']} "
                if (
                    hw_pc := self._model.pulse_channel_with_id(pulse_channel_id)
                ) is not None:
                    pc_name = f"{hw_pc.pulse_type}"
                else:
                    pc_name = f"Custom({pulse_channel_id})"

                freq_range = self._get_if_freq_range(physical_channel_data["is_resonator"])
                violations.append(
                    f"The IF of {name} {pc_name} pulse channel has a value {if_freq}, "
                    f"which is outside of the the valid range {freq_range}."
                )

        return violations

    def run(
        self, ir: QuantumInstructionBuilder, res_mgr: ResultManager, *args, **kwargs
    ) -> QuantumInstructionBuilder:
        """
        :param ir: The list of instructions stored in an :class:`QuantumInstructionBuilder`.
        :param res_mgr: The result manager containing the results of the analysis.
        """

        active_channel_res = res_mgr.lookup_by_type(ActivePulseChannelResults)
        pulse_channels = active_channel_res.targets
        physical_channels = set(
            ir.get_pulse_channel(pulse_channel).physical_channel_id
            for pulse_channel in pulse_channels
        )

        violations = self._find_physical_channel_violations(physical_channels)
        violations.extend(self._find_pulse_channel_violations(pulse_channels, ir))

        if len(violations) > 0:
            raise ValueError(
                "Frequency validation of the hardware model against the target data failed "
                "with the following violations: \n" + "\n".join(violations)
            )
        return ir


class ReturnSanitisationValidation(ValidationPass):
    """Validates that the IR has a :class:`Return` instruction."""

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        returns = [inst for inst in ir.instructions if isinstance(inst, Return)]

        if not returns:
            raise ValueError("Could not find any return instructions.")
        elif len(returns) > 1:
            raise ValueError("Found multiple return instructions.")
        return ir


class RepeatSanitisationValidation(ValidationPass):
    """Checks if the builder has a :class:`Repeat` instruction and warns if none exists."""

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""
        for inst in ir.instructions:
            if isinstance(inst, Repeat):
                return ir

        log.warning("Could not find any repeat instructions.")
        return ir


PydInstructionValidation = InstructionValidation
PydReadoutValidation = ReadoutValidation
PydDynamicFrequencyValidation = DynamicFrequencyValidation
PydHardwareConfigValidity = HardwareConfigValidity
PydFrequencySetupValidation = FrequencySetupValidation
PydNoMidCircuitMeasurementValidation = NoMidCircuitMeasurementValidation
PydReturnSanitisationValidation = ReturnSanitisationValidation
PydRepeatSanitisationValidation = RepeatSanitisationValidation



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import csv
import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Generic, TypeVar

from pydantic import BaseModel, ConfigDict, Field, IPvAnyAddress

from qat.engines import ConnectionMixin
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class InstrumentModel(BaseModel):
    """
    Used to parse JSON/CSV entries. An instrument has an id, name, and IP address
    """

    model_config = ConfigDict(populate_by_name=True)

    id: str = Field(frozen=True, alias="ID")
    name: str = Field(frozen=True, alias="NAME")
    address: IPvAnyAddress = Field(frozen=True, alias="ADDRESS")


class InstrumentConcept(ConnectionMixin):
    """
    The component interface, it defines what an instrument is and what APIs through which
    customer code can interact with.
    """

    def connect(self):
        pass

    def disconnect(self):
        pass

    def setup(self, *args, **kwargs):
        pass

    def playback(self, *args, **kwargs):
        pass


class LeafInstrument(InstrumentConcept):
    def __init__(self, id: str, name: str, address: str):
        self.id = id
        self.name = name
        self.address = address

        self._is_connected = False

    @property
    def is_connected(self) -> bool:
        return self._is_connected

    @is_connected.setter
    def is_connected(self, value: bool):
        self._is_connected = value

    def __repr__(self):
        return f"{self.id}_{self.name}_{self.address}"

    def __str__(self):
        return self.__repr__()


LInstr = TypeVar("LInstr", bound=LeafInstrument)


class CompositeInstrument(Generic[LInstr], InstrumentConcept):
    def __init__(self):
        self._components: dict[str, LInstr] = {}

    @property
    def components(self) -> dict[str, LInstr]:
        return self._components

    @property
    def is_connected(self) -> bool:
        return all((component.is_connected for component in self._components.values()))

    def connect(self):
        for comp in self._components.values():
            comp.connect()

    def disconnect(self):
        for comp in self._components.values():
            comp.disconnect()

    def add(self, component: LInstr):
        if next(
            (comp for id, comp in self._components.items() if id == component.id),
            None,
        ):
            raise ValueError(f"Instrument {component} already exists")
        else:
            self._components[component.id] = component

    def setup(self, *args, **kwargs):
        for comp in self._components.values():
            comp.setup(*args, **kwargs)

    def playback(self, *args, **kwargs):
        component_results = [
            comp.playback(*args, **kwargs) for comp in self._components.values()
        ]

        results = {}
        for comp_res in component_results:
            if comp_res.keys() & results.keys():
                raise ValueError("Component results have conflicting keys")
            results.update(comp_res)

        return results

    def __repr__(self):
        return "\n".join(repr(comp) for comp in self._components.values())

    def __str__(self):
        return "\n".join(str(comp) for comp in self._components.values())


CInstr = TypeVar("CInstr", bound=CompositeInstrument)


class InstrumentBuilder(ABC):
    @abstractmethod
    def build(self, *args, **kwargs) -> CInstr: ...


class CsvInstrumentBuilder(InstrumentBuilder):
    def __init__(
        self, file_path: str | Path, cinstr_type: type = None, linstr_type: type = None
    ):
        self.file_path = file_path
        self.cinstr_type = cinstr_type or CompositeInstrument
        self.linstr_type = linstr_type or LeafInstrument

    def build(self) -> CInstr:
        """
        A generic function that builds an InstrumentConcept object representing
        an arbitrary fleet of leaf instruments defined as CSV.

        :param cinstr_type: type of Composite Instrument to build
        :param linstr_type: type of Leaf Instrument to build
        """

        if not os.path.exists(self.file_path):
            raise ValueError(f"File '{self.file_path}' not found!")

        composite = self.cinstr_type()
        with open(self.file_path) as f:
            reader = csv.DictReader(f)
            for row in reader:
                instr_model = InstrumentModel.model_validate(row)
                composite.add(
                    self.linstr_type(
                        id=instr_model.id,
                        name=instr_model.name,
                        address=str(instr_model.address),
                    )
                )

        return composite


class ConfigInstrumentBuilder(InstrumentBuilder):
    """
    Builds an InstrumentConcept object representing an arbitrary fleet of leaf instruments
    defined as a list of config dicts.

    :param cinstr_type: type of Composite Instrument to build
    :param linstr_type: type of Leaf Instrument to build
    """

    def __init__(
        self, configs: list[dict], cinstr_type: type = None, linstr_type: type = None
    ):
        self.configs = configs
        self.cinstr_type = cinstr_type or CompositeInstrument
        self.linstr_type = linstr_type or LeafInstrument

    def build(self) -> CInstr:
        composite = self.cinstr_type()
        for config in self.configs:
            instr_model = InstrumentModel.model_validate(config)
            composite.add(
                self.linstr_type(
                    id=instr_model.id,
                    name=instr_model.name,
                    address=str(instr_model.address),
                )
            )

        return composite



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import numpy as np

from qat.core.pass_base import TransformPass
from qat.ir.gates.base import GateBase, QubitInstruction
from qat.ir.gates.gates_1q import Gate1Q, Phase, Rx, Rz, U, X, Z
from qat.ir.gates.gates_2q import CNOT, ECR, Gate2Q
from qat.ir.gates.native import X_pi_2
from qat.ir.qat_ir import QatIR
from qat.middleend.decompositions.gates import DefaultGateDecompositions
from qat.middleend.decompositions.pulses import (
    DefaultPulseDecompositions,
    PulseDecompositionBase,
)
from qat.model.hardware_model import PhysicalHardwareModel


class Decompose2QToCNOTs(TransformPass):
    r"""Iterates through the list of gates and decomposes any two-qubit gate down to CNOTs.

    This applies to any 2Q gate with the exception of ECR, :math:`ZX(\pi/4)` and
    :math:`ZX(-\pi/4)`, which sit lower in the DAG than CNOTs. The benefit of this pass is
    that CNOTs have relatively simple ir algebra and allow for easy optimisations. This
    might later be extended to include ZX-calculus techniques. An example of this is
    exploiting the fact that a CNOT is involutory, meaning we can cancel two CNOTs that
    appear sequentially.
    """

    def __init__(self):
        self.decompositions = DefaultGateDecompositions(extend_end_nodes=[CNOT, ECR])

    def run(self, ir: QatIR, *args, **kwargs):
        """:param ir: QatIR to transform."""
        new_instructions = []
        for gate in ir.instructions:
            if isinstance(gate, Gate2Q):
                new_instructions.extend(self.decompositions.decompose(gate))
            else:
                new_instructions.append(gate)
        ir.instructions = new_instructions
        return ir


class DecomposeToNativeGates(TransformPass):
    r"""Iterates through the list of gates and decomposes gate into a native gate set.

    Default native gates are :class:`Z`, :class:`X_pi_2` and :class:`ZX_pi_4`.
    """

    def __init__(self, native_gates: list[GateBase] | None = None):
        """
        :param native_gates: A list of native gates to decompose to.
        """
        self.decompositions = DefaultGateDecompositions(end_nodes=native_gates)

    def run(self, ir: QatIR, *args, **kwargs):
        """:param ir: QatIR to transform."""
        new_instructions = []
        for gate in ir.instructions:
            new_instructions.extend(self.decompositions.decompose(gate))
        ir.instructions = new_instructions
        return ir


class DecomposeToPulses(TransformPass):
    """A simple pass proof-of-concept pass that decomposes gates to pulse instructions.

    Each native gate can be written a sequence of pulse instructions. This pass decomposes
    each gate into the correct instructions. Note that this method will not be supported in
    the long-term, and will be replaced by something more performant once Qat IR is
    better established.

    A simple first step might be to write an analysis pass that extracts a set of unique
    gates + qubit(s) targets. For each gate+target pair, it could then determine the set of
    instructions for pair just once. There is the complication of how to deal with the
    Rz-gate, which is parameterised. This pass could then be adapted to use the results of
    this pass to more effeciently decompose instructions.
    """

    def __init__(
        self, model: PhysicalHardwareModel, decompositions: PulseDecompositionBase = None
    ):
        """Initiate the pass with the hardware model.

        :param model: The hardware model is needed to construct the pulses.
        """
        self.model = model
        if not decompositions:
            decompositions = DefaultPulseDecompositions()
        self.decompositions = decompositions

    def run(self, ir: QatIR, *args, **kwargs):
        """:param ir: The IR containing gates."""

        new_instructions = []
        for inst in ir.instructions:
            if not isinstance(inst, QubitInstruction):
                new_instructions.append(inst)
                continue
            new_instructions.extend(self.decompositions.decompose(inst, self.model))

        ir.instructions = new_instructions
        return ir


class Squash1QGates(TransformPass):
    r"""Squashes consecutive 1Q qubits into a single :math:`U(\theta, \phi, \lambda)` gate.

    Iterates through the list of instructions, looking for series of 1Q gates that are not
    interupted by any non-1Q gates. It then squashes these gates into a single U gate, which
    will later be composed into native gates.
    """

    def run(self, ir: QatIR, *args, **kwargs):
        """:param ir: QatIR to transform."""
        gate_sequences: dict[int, list[int]] = {}
        new_instructions: list[GateBase] = []
        for idx, gate in enumerate(ir.instructions):
            if isinstance(gate, Gate1Q):
                gate_sequences.setdefault(gate.qubit, []).append(idx)
            else:
                for qubit in gate.qubits:
                    seq = gate_sequences.get(qubit, [])
                    if len(seq) == 0:
                        pass
                    else:
                        new_instructions.append(
                            self.squash_gate([ir.instructions[i] for i in seq])
                        )
                        del gate_sequences[qubit]
                new_instructions.append(gate)
        for seq in gate_sequences.values():
            new_instructions.append(self.squash_gate([ir.instructions[i] for i in seq]))
        ir.instructions = new_instructions
        return ir

    def squash_gate(self, gates: list[Gate1Q]):
        """Squashes a list of gates into a U gate."""
        # can probably be moved elsewhere for reuse.
        if len(gates) == 1:
            return gates[0]
        mat = np.eye(2)
        for gate in gates:
            mat = gate.matrix @ mat
        return U.from_matrix(gates[0].qubit, mat)


class Squash2QGates(TransformPass):
    r"""Squashes consecutive 1Q and 2Q qubits into a single gate.

    Iterates through the list of instructions, looking for consecutive 1Q and 2Q gates that
    act on the same pair of qubits. It then squashes these gates into a single U gate, which
    will later be composed into native gates.
    """

    def run(self, ir: QatIR, *args, **kwargs):
        raise NotImplementedError(
            "Optimisation pass not yet implemented. Requires implementation of general 2Q "
            "decompositions."
        )


class SquashCNOTs(TransformPass):
    r"""Iterates through a list of gates and removes redundant CNOTs.

    If two CNOTs that share the same target and control qubits appear sequentially [note
    that this means with respect to these two qubits, and not other qubits], then they can
    be squashed to an identity matrix. This also exploits the fact :math:`R_{z}(\theta)`
    gates that act on the control qubit and :math:`R_{x}(\theta)` gates that act on target
    qubits can be commuted through CNOTs, allowing CNOTs to be cancelled.

    This can be considered a simple optimisation using ZX-calculus: in the future, we can
    build on this more.
    """

    def run(self, ir: QatIR, *args, **kwargs):
        """:param ir: QatIR to transform."""
        # find all cnots and group them by control-target pairs
        cnots = [
            (pos, gate.qubits)
            for pos, gate in enumerate(ir.instructions)
            if isinstance(gate, CNOT)
        ]
        unique_pairs = set([cnot[1] for cnot in cnots])
        qubits_cnot_map = {
            pair: [cnot[0] for cnot in cnots if cnot[1] == pair] for pair in unique_pairs
        }

        # Find which CNOTs can be squashed by checking if CNOTs are sequential for those
        # qubits
        del_positions = []
        for pair, positions in qubits_cnot_map.items():
            if len(positions) == 1:
                continue
            pair_set = set(pair)
            intersections = [
                len(set(gate.qubits) & pair_set) == 0
                or (isinstance(gate, (Z, Rz, Phase)) and gate.qubit == pair[0])
                or (isinstance(gate, (X, Rx, X_pi_2)) and gate.qubit == pair[1])
                for gate in ir.instructions[positions[0] : positions[-1]]
            ]
            i = 0
            while i < len(positions) - 1:
                start_pos = positions[i] - positions[0] + 1
                end_pos = positions[i + 1] - positions[0]
                if all(intersections[start_pos:end_pos]):
                    del_positions.append(positions[i])
                    del_positions.append(positions[i + 1])
                    i += 2
                else:
                    i += 1

        # remove squashed CNOTs
        del_positions = sorted(del_positions, reverse=True)
        for pos in del_positions:
            del ir.instructions[pos]
        return ir


class SyncEndMeasurements(TransformPass):
    """Finds and synchronizes all end-circuit measurements to happen at the same time.

    Bubbles end-measurements to the end of the instruction list, and adds a barrier before
    them.
    """

    def run(self, ir: QatIR, *args, **kwargs):
        raise NotImplementedError(
            "Transformation passes for the end measurements are not implemented yet."
        )


class SequentialEndMeasurements(TransformPass):
    """Finds all end measurements and enforces that they happen at different times.

    Bubbles end-measurements to the end of the instruction list, and adds a barrier between
    them.
    """

    def run(self, ir: QatIR, *args, **kwargs):
        raise NotImplementedError(
            "Transformation passes for the end measurements are not implemented yet."
        )


class StaggeredEndMeasurements(TransformPass):
    """Finds all end measurements, and staggers their timings so that no Qubits with
    couplings have measurements occur at the same time.

    Bubbles end-measurements to the end of the instruction list. Some analysis is required
    to effeciently bundle together non-coupled Qubits.
    """

    def run(self, ir: QatIR, *args, **kwargs):
        raise NotImplementedError(
            "Transformation passes for the end measurements are not implemented yet."
        )


class RemoveRedundantSWAPs(TransformPass):
    """Indentifies redundant SWAPs at the beginning and the end of a circuit, and removes
    them. This pass is designed to be used before routing.

    SWAP gates can frequently be found in circuits to allow for multi-qubit gates to be
    applied on qubits that are not connected in the hardware's topology. However, SWAP gates
    are expensive to implement (requiring three CNOT gates), and sometimes the same circuit
    can be implemented through a smart choice of logical-to-physical qubit placement.

    .. warning::
        This pass can modify a circuit that is capable of running on the given topology
        to one that might fail due to gates applied on non-connected qubits. The intention
        is that this is done before the routing and placement methods are applied to
        potentially reduce the overall 2Q-gate count.
    """

    def run(self, ir: QatIR, *args, **kwargs):
        raise NotImplementedError("This transformation pass is not yet implemented.")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from abc import abstractmethod
from typing import get_type_hints

from pydantic import BaseModel, TypeAdapter

from qat.engines.model import requires_hardware_model
from qat.engines.native import NativeEngine
from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.loaders.base import BaseModelLoader
from qat.model.target_data import TargetData
from qat.model.validators import MismatchingHardwareModelException
from qat.pipelines.base import AbstractPipeline
from qat.pipelines.pipeline import BasePipeline
from qat.purr.compiler.hardware_models import QuantumHardwareModel


class PipelineConfig(BaseModel):
    """Base class for configuring updateable pipelines. Subclasses of
    :class:`UpdateablePipeline` should be paried with their own configuration class which
    specifies custom configuration parameters, and/or sets custom defaults."""

    name: str


Model = QuantumHardwareModel | PhysicalHardwareModel


class UpdateablePipeline(AbstractPipeline):
    """A factory for creating and updating pipelines.

    Pipelines are designed to be immutable to maintain performance and prevent errors from
    moving parts. :class:`UpdateablePipeline` provide a mechanism to update and refresh
    pipelines with updated components, such as an updated hardware model, without
    invalidating other components in the pipeline by reconstructing the entire pipeline.
    It will also inherit the properties and methods of the pipeline it creates, allowing
    it to be used as a drop-in replacement for the original pipeline.

    :class:`UpdateablePipeline` can be instantiated with either a hardware model or a
    loader, or both. If both are provided, the model will take precedence for the initial
    pipeline constuction, but the loader can be used to refresh the model later on. If a
    loader is provided but not a model, the loader will be used to load the model
    during the initial pipeline construction.
    """

    def __init__(
        self,
        config: PipelineConfig,
        model: Model | None = None,
        loader: BaseModelLoader | None = None,
        target_data: TargetData | None = None,
        engine: NativeEngine | None = None,
    ):
        """
        :param config: The pipeline configuration with the name of the pipeline, and any
            additional parameters that can be configured in the pipeline.
        :param model: The hardware model to feed into the pipeline. Defaults to None.
        :param loader: The hardware loader used to load the hardware model which can be used
            to later refresh the hardware model. Defaults to None.
        :param target_data: The data concerning the target device, defaults to None
        :param engine: The engine to use for the pipeline, defaults to None.
        :raises ValueError: If neither model nor loader is provided.
        """

        config = TypeAdapter(self.config_type()).validate_python(config)

        if model is None and loader is None:
            raise ValueError("Model or loader must be provided.")

        if model is None:
            model = loader.load()

        # TODO: Change to just check against RequiresHardwareModelMixin (COMPILER-662)
        if requires_hardware_model(engine):
            if engine.model != model:
                raise MismatchingHardwareModelException(
                    f"Engine model {engine.model} does not match the provided model "
                    f"{model}. If the UpdateablePipeline is only provided a loader, and "
                    "not a model, please consider also instantiating with a model."
                )

        self._loader = loader
        self._engine = engine
        self._pipeline_config = config
        self._pipeline = self.__class__._build_pipeline(config, model, target_data, engine)

    @staticmethod
    @abstractmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: Model,
        target_data: TargetData | None,
        engine: NativeEngine | None = None,
    ) -> BasePipeline:
        """Build the pipeline based on the configuration."""

        raise NotImplementedError(
            "Custom pipelines that are subclasses of the `UpdateablePipeline` are required "
            "to implement their own `_build_pipeline` method that returns a `Pipeline` "
            "object."
        )

    def is_subtype_of(self, cls):
        """Matches the type against the pipeline produced by the factory."""
        return isinstance(self, cls) or self._pipeline.is_subtype_of(cls)

    @classmethod
    def config_type(cls) -> type:
        """Inspects the type of the config attribute in :meth:`_build_pipeline`."""
        type_hints = get_type_hints(cls._build_pipeline)
        if "config" not in type_hints:
            return PipelineConfig
        return get_type_hints(cls._build_pipeline)["config"]

    @property
    def name(self) -> str:
        """Return the name of the pipeline."""
        return self._pipeline_config.name

    @property
    def model(self) -> Model:
        """Return the model associated with the pipeline."""
        return self._pipeline.model

    @model.setter
    def model(self, value: Model):
        """Updates the pipeline with some set model."""
        self.update(model=value)

    @property
    def target_data(self) -> TargetData | None:
        """Return the target data associated with the pipeline."""
        return self._pipeline.target_data

    @property
    def pipeline(self) -> BasePipeline:
        """Return the pipeline instance."""
        return self._pipeline

    @property
    def config(self) -> PipelineConfig:
        """Return the pipeline configuration."""
        return self._pipeline_config

    @property
    def has_loader(self) -> bool:
        """Check if the pipeline has a loader."""
        return self._loader is not None

    def update(
        self,
        config: PipelineConfig | None = None,
        model: Model | None = None,
        loader: BaseModelLoader | None = None,
        target_data: TargetData | None = None,
        engine: NativeEngine | None = None,
        reload_model: bool = False,
    ):
        """Update the pipeline configuration and rebuild the pipeline with updated
        arguments. The whole pipeline is reinstantiated to avoid conflicts with changing
        components."""

        model, self._loader, reload_model = self._resolve_model(model, loader, reload_model)

        if config is not None:
            config = TypeAdapter(self.config_type()).validate_python(config)
            if (config.name is not None) and config.name != self._pipeline_config.name:
                raise ValueError("Cannot change the name of the pipeline during an update.")

            # Update the pipeline configuration with the new config
            self._pipeline_config = self._pipeline_config.model_copy(
                update=config.model_dump(exclude_unset=True, exclude_defaults=True)
            )

        if target_data is None:
            target_data = self.target_data

        engine = self._engine if engine is None else engine
        # TODO: Change to just check against RequiresHardwareModelMixin (COMPILER-662)
        if requires_hardware_model(engine) and reload_model:
            engine.model = model
            self._engine = engine

        self._pipeline = self.__class__._build_pipeline(
            config=self._pipeline_config,
            model=model,
            target_data=target_data,
            engine=engine,
        )

    def copy(self) -> "UpdateablePipeline":
        """Create a copy of the pipeline factory."""
        return self.__class__(
            config=self._pipeline_config,
            model=self.model,
            loader=self._loader,
            target_data=self.target_data,
            engine=self.engine,
        )

    def copy_with(
        self,
        config: PipelineConfig | None = None,
        model: Model | None = None,
        loader: BaseModelLoader | None = None,
        target_data: TargetData | None = None,
        engine: NativeEngine | None = None,
        reload_model: bool = False,
    ) -> "UpdateablePipeline":
        """Create a copy of the pipeline factory with updated parameters."""

        model, loader, reload_model = self._resolve_model(model, loader, reload_model)

        if config is not None:
            config = TypeAdapter(self.config_type()).validate_python(config)
            config = self._pipeline_config.model_copy(
                update=config.model_dump(exclude_unset=True, exclude_defaults=True)
            )

        engine = self._engine if engine is None else engine
        # TODO: Change to just check against RequiresHardwareModelMixin (COMPILER-662)
        if requires_hardware_model(engine) and reload_model:
            engine.model = model

        return self.__class__(
            config=config if config is not None else self._pipeline_config,
            model=model,
            loader=loader,
            target_data=target_data if target_data is not None else self.target_data,
            engine=engine,
        )

    def _resolve_model(
        self, model: Model, loader: BaseModelLoader | None, reload_model: bool
    ) -> tuple[Model, BaseModelLoader | None, bool]:
        """Resolves the model from the model, loader and reload_model parameters.

        The model is returned, along with the loader and a boolean indicating if the model
        is updated.
        """

        if model is not None and reload_model:
            raise ValueError("Both model and reload_model cannot be used together.")

        if model is None and loader is None and not reload_model:
            return self._pipeline.model, self._loader, False

        if loader is not None and model is None:
            reload_model = True

        if loader is None:
            loader = self._loader

        if reload_model:
            if loader is None:
                raise ValueError("Factory has no Model loader, cannot reload model.")
            model = loader.load()

        return model, loader, True

    def __getattr__(self, name: str):
        """Delegate attribute access to the underlying pipeline instance."""

        if name in self.__dict__:
            return self.__dict__[name]
        if hasattr(self._pipeline, name):
            return getattr(self._pipeline, name)
        raise AttributeError(f"{self.__class__.__name__} has no attribute '{name}'")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from typing import Literal, Type

from compiler_config.config import CompilerConfig
from pydantic import BaseModel, ConfigDict, Field, field_validator
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    YamlConfigSettingsSource,
)
from qiskit_aer import AerSimulator

from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class QiskitSimulationConfig(BaseModel):
    """
    The default settings for the Qiskit Simulator, including overridden MPS settings.
    """

    model_config = ConfigDict(validate_assignment=True)
    _allowed_methods: type = Literal[AerSimulator().available_methods()]
    METHOD: _allowed_methods = "automatic"
    """The simulation method to use."""
    FALLBACK_SEQUENCE: list[_allowed_methods] = ["automatic", "matrix_product_state"]
    """If the simulation fails, specify a fallback sequence of methods to call."""
    OPTIONS: dict = {
        "matrix_product_state_max_bond_dimension": 128,
        "matrix_product_state_truncation_threshold": 1e-12,
    }
    """
    Specify the options for a chosen AerSimulator backend. See
    https://docs.quantum.ibm.com/api/qiskit/0.37/qiskit.providers.aer.AerSimulator
    for options you can provide.
    """
    ENABLE_METADATA: bool = False
    """Returns the AerSimulator metadata if enabled."""


class QatSimulationConfig(BaseModel):
    """
    The default settings for QATs simulation backends.
    """

    model_config = ConfigDict(validate_assignment=True)
    QISKIT: QiskitSimulationConfig = QiskitSimulationConfig()


class InstructionValidationConfig(BaseModel):
    """
    The default settings for validation of instructions.
    """

    model_config = ConfigDict(validate_assignment=True)
    NO_MID_CIRCUIT_MEASUREMENT: bool = True
    MAX_INSTRUCTION_LENGTH: bool = True
    ACQUIRE_CHANNEL: bool = True
    PULSE_DURATION_LIMITS: bool = True

    @property
    def DISABLED(self):
        return not (
            self.NO_MID_CIRCUIT_MEASUREMENT
            | self.MAX_INSTRUCTION_LENGTH
            | self.ACQUIRE_CHANNEL
            | self.PULSE_DURATION_LIMITS
        )

    def disable(self):
        self.NO_MID_CIRCUIT_MEASUREMENT = False
        self.MAX_INSTRUCTION_LENGTH = False
        self.ACQUIRE_CHANNEL = False
        self.PULSE_DURATION_LIMITS = False

    @field_validator("PULSE_DURATION_LIMITS")
    def check_disable_pulse_duration_limits(cls, PULSE_DURATION_LIMITS):
        if not PULSE_DURATION_LIMITS:
            log.warning(
                "Disabled check for pulse duration limits, which should ideally only be used for calibration purposes."
            )
        return PULSE_DURATION_LIMITS


class QatConfig(BaseSettings):
    """
    Full settings for a single job. Allows environment variables to be overridden by direct assignment.
    """

    model_config = SettingsConfigDict(
        env_prefix="QAT_",
        env_nested_delimiter="_",
        validate_assignment=True,
    )
    MAX_REPEATS_LIMIT: int | None = Field(gt=0, default=100_000)
    """Max number of repeats / shots to be performed in a single job."""

    INSTRUCTION_VALIDATION: InstructionValidationConfig = InstructionValidationConfig()
    """Options for Instruction validation before execution."""

    SIMULATION: QatSimulationConfig = QatSimulationConfig()
    """Options for QATs simulation backends."""

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            env_settings,
            YamlConfigSettingsSource(settings_cls),
        )

    def validate(self, compiler_config: CompilerConfig):
        """_summary_

        Args:
            compiler_config (CompilerConfig): _description_
        """
        if compiler_config.repeats and (compiler_config.repeats > self.MAX_REPEATS_LIMIT):
            raise ValueError(
                f"Number of shots {compiler_config.repeats} exceeds the maximum amount of {self.MAX_REPEATS_LIMIT}."
            )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from collections import defaultdict

import numpy as np

from qat.backend.base import BaseBackend
from qat.backend.passes.analysis import TimelineAnalysis, TimelineAnalysisResult
from qat.backend.passes.lowering import PartitionByPulseChannel
from qat.backend.passes.validation import NoAcquireWeightsValidation
from qat.backend.waveform.executable import (
    PositionalAcquireData,
    WaveformChannelData,
    WaveformProgram,
)
from qat.core.pass_base import InvokerMixin, MetricsManager, PassManager
from qat.core.result_base import ResultManager
from qat.executables import AcquireData, Executable
from qat.ir.instructions import FrequencyShift, PhaseReset, PhaseSet, PhaseShift, Reset
from qat.ir.lowered import PartitionedIR
from qat.ir.waveforms import Pulse, SampledWaveform
from qat.model.device import PhysicalChannel, Qubit
from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.target_data import DeviceDescription, TargetData
from qat.purr.backends.utilities import UPCONVERT_SIGN
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PulseChannel
from qat.purr.compiler.instructions import Instruction
from qat.purr.utils.logger import get_default_logger
from qat.utils.waveform import NumericFunction

log = get_default_logger()


class WaveformBackend(BaseBackend[WaveformProgram], InvokerMixin):
    """
    Target-machine code generation from an IR for targets that only require the explicit waveforms.
    """

    def __init__(
        self,
        model: PhysicalHardwareModel,
        target_data: TargetData = TargetData.default(),
    ):
        self.model = model
        self.target_data = target_data

    def emit(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        upconvert: bool = True,
        **kwargs,
    ) -> Executable[WaveformProgram]:
        """Compiles :class:`InstructionBuilder` into an :class:`Executable`.

        Translates pulse instructions into explicit waveforms at the required times, and
        combines them across pulse channels to give a composite waveform on the necessary
        physical channels. Deals with low-level details such as pulse scheduling and phase
        shifts.
        """
        res_mgr = res_mgr if res_mgr is not None else ResultManager()
        met_mgr = met_mgr if met_mgr is not None else MetricsManager()
        ir = self.run_pass_pipeline(ir, res_mgr, met_mgr, **kwargs)
        timeline_res = res_mgr.lookup_by_type(TimelineAnalysisResult)

        buffers = self.create_physical_channel_buffers(ir, timeline_res, upconvert)
        execute_acquire_dict, runtime_acquire_dict = self.create_acquires(ir, timeline_res)

        channels: dict[str, WaveformChannelData] = {}
        for physical_channel, buffer in buffers.items():
            channels[physical_channel.uuid] = WaveformChannelData(
                buffer=buffer,
                acquires=execute_acquire_dict.get(physical_channel, []),
            )

        returns = []
        for _return in ir.returns:
            returns.extend(_return.variables)

        programs = self.create_programs(
            channels,
            ir.shots,
            ir.compiled_shots,
            timeline_res.total_duration,
        )

        return Executable[WaveformProgram](
            programs=programs,
            acquires=runtime_acquire_dict,
            assigns=ir.assigns,
            returns=returns,
            calibration_id=self.model.calibration_id,
            shots=ir.shots,
        )

    def build_pass_pipeline(self, *args, **kwargs):
        return (
            PassManager()
            | NoAcquireWeightsValidation()
            | PartitionByPulseChannel()
            | TimelineAnalysis(self.model, self.target_data)
        )

    @staticmethod
    def create_pulse_channel_buffer(
        pulse_channel: PulseChannel,
        instructions: list[Instruction],
        durations: list[int],
        device_description: DeviceDescription,
        phys_channel: PhysicalChannel,
        upconvert: bool = True,
    ):
        """Creates a buffer of waveforms for a single pulse channel.

        :param pulse_channel: The pulse channel to create the buffer for.
        :param instructions: The list of instructions to process.
        :param durations: The list of durations for each instruction.
        :param upconvert: Whether to upconvert the waveforms to the target frequencies.
        """
        context = WaveformContext(
            pulse_channel, np.sum(durations), device_description, phys_channel
        )
        for i, instruction in enumerate(instructions):
            duration = durations[i]
            if isinstance(instruction, Pulse):
                context.process_pulse(instruction, duration, upconvert)
            elif isinstance(instruction, PhaseShift):
                context.process_phaseshift(instruction.phase)
            elif isinstance(instruction, PhaseSet):
                context.process_phaseset(instruction.phase)
            elif isinstance(instruction, PhaseReset):
                context.process_phasereset()
            elif isinstance(instruction, Reset):
                context.process_reset()
            elif isinstance(instruction, FrequencyShift):
                context.process_frequencyshift(instruction.frequency)
            else:
                context.process_delay(duration)
        return context.buffer

    def create_physical_channel_buffers(
        self,
        ir: PartitionedIR,
        timeline_res: TimelineAnalysisResult,
        upconvert: bool = True,
    ):
        """Creates a buffer of waveforms for each physical channel.

        :param ir: The partitioned IR containing the instructions.
        :param timeline_res: The timeline analysis result containing the number of samples
            for each instruction.
        :param upconvert: Whether to upconvert the waveforms to the target frequencies.
        """
        # Compute all pulse channel buffers
        pulse_channel_buffers = {}
        for pulse_channel_id in timeline_res.target_map:
            # The buffer is trivial if there are no waveforms
            if not any(
                [isinstance(inst, Pulse) for inst in ir.target_map[pulse_channel_id]]
            ):
                continue

            pulse_channel = ir.get_pulse_channel(pulse_channel_id)
            device = self.model.device_for_physical_channel_id(
                pulse_channel.physical_channel_id
            )
            pulse_channel_buffers[pulse_channel_id] = self.create_pulse_channel_buffer(
                pulse_channel,
                ir.target_map[pulse_channel_id],
                timeline_res.target_map[pulse_channel_id].samples,
                self.target_data.QUBIT_DATA
                if isinstance(device, Qubit)
                else self.target_data.RESONATOR_DATA,
                self.model.physical_channel_with_id(pulse_channel.physical_channel_id),
                upconvert,
            )

        # Organize by physical channels
        phys_to_pulse_map = {}
        for pulse_channel_id, buffer in pulse_channel_buffers.items():
            physical_channel = self.model.physical_channel_with_id(
                ir.get_pulse_channel(pulse_channel_id).physical_channel_id
            )
            phys_to_pulse_map.setdefault(physical_channel, []).append(buffer)

        # Compute sum of pulse channels
        physical_channel_buffers = {}
        for device in self.model.quantum_devices:
            physical_channel = device.physical_channel
            buffer_list = phys_to_pulse_map.get(physical_channel, [])
            if len(buffer_list) == 0:
                length = 0
            else:
                length = max(0, *[len(buffer) for buffer in buffer_list])
            buffer_sum = np.zeros(length, dtype=np.complex128)
            for buffer in buffer_list:
                buffer_sum[0 : len(buffer)] += buffer
            physical_channel_buffers[physical_channel] = buffer_sum

        return physical_channel_buffers

    def create_acquires(self, ir: PartitionedIR, timeline_res: TimelineAnalysisResult):
        """Assembles the acquire data needed to perform acquisition, and also the
        acquire information needed at runtime.

        :param ir: The partitioned IR containing the acquire map.
        :param timeline_res: The timeline analysis result containing the number of samples
            for each instruction.
        """
        execute_acquire_dict = defaultdict(list)
        runtime_acquire_dict = dict()
        for pulse_channel_id, acquire_list in ir.acquire_map.items():
            pulse_channel = ir.get_pulse_channel(pulse_channel_id)
            phys_channel = self.model.physical_channel_with_id(
                pulse_channel.physical_channel_id
            )
            acq_list = execute_acquire_dict[phys_channel]
            for acquire in acquire_list:
                idx = ir.target_map[pulse_channel_id].index(acquire)
                acq_list.append(
                    PositionalAcquireData(
                        length=timeline_res.target_map[pulse_channel_id].samples[idx],
                        position=np.sum(
                            timeline_res.target_map[pulse_channel_id].samples[0:idx]
                        ),
                        mode=acquire.mode,
                        output_variable=acquire.output_variable,
                    )
                )

                rp = ir.rp_map.get(acquire.output_variable, None)
                rp = rp.results_processing if rp is not None else None
                runtime_acquire_dict[acquire.output_variable] = AcquireData(
                    mode=acquire.mode,
                    post_processing=ir.pp_map.get(acquire.output_variable, []),
                    shape=(ir.shots,),
                    results_processing=rp,
                    physical_channel=phys_channel.uuid,
                )

        return execute_acquire_dict, runtime_acquire_dict

    def create_programs(
        self,
        channel_data: dict[str, WaveformChannelData],
        shots: int,
        batch_size: int,
        repetition_time: float,
    ) -> WaveformProgram:
        """Creates WaveformProgram instances from channel data.

        :param channel_data: The channel data for each physical channel.
        :param shots: The total number of shots to be executed.
        :param batch_size: The maximum number of shots per program.
        :param repetition_time: The time for each shot.
        """
        program = WaveformProgram(
            channel_data=channel_data, repetition_time=repetition_time, shots=batch_size
        )
        programs = [program for _ in range(shots // batch_size)]
        if (remaining_shots := shots % batch_size) > 0:
            programs.append(
                WaveformProgram(
                    channel_data=channel_data,
                    repetition_time=repetition_time,
                    shots=remaining_shots,
                )
            )
        return programs


class WaveformContext:
    """Used for waveform code generation for a particular pulse channel.

    This can be considered to be the dynamical state of a pulse channel which evolves as the
    circuit progresses.
    """

    # TODO: refactor this as a "channel backend" or as passes depending on when this is
    # done (COMPILER-413).

    def __init__(
        self,
        pulse_channel: PulseChannel,
        total_duration: int,
        device_description: DeviceDescription,
        phys_channel: PhysicalChannel,
    ):
        """
        :param pulse_channel: The pulse channel that this is modelling.
        :param total_duration: The lifetime of the pulse channel in number of samples.
        """
        self.pulse_channel = pulse_channel
        self.physical_channel = phys_channel
        self._device_desc = device_description
        self._buffer = np.zeros(total_duration, dtype=np.complex128)
        self._duration = 0
        self._phase = 0.0
        self._frequency = pulse_channel.frequency

    @property
    def buffer(self):
        return self._buffer

    def process_pulse(self, instruction: Pulse, samples: int, do_upconvert: bool = True):
        """Converts a waveform instruction into a discrete number of samples, handling
        upconversion to the target frequency if specified."""

        # TODO: the evaluate shape is handled in the EvaluatePulses pass, so this needs
        # adjusting to only accept square waveforms and custom pulses. (COMPILER-413)

        dt = self._device_desc.sample_time
        length = samples * dt
        centre = length / 2.0
        t = np.linspace(
            start=-centre + 0.5 * dt, stop=length - centre - 0.5 * dt, num=samples
        )
        waveform = instruction.waveform

        if isinstance(waveform, SampledWaveform):
            amp = 1.0
            scale_factor = 1.0
            drag = 0.0
            amplitude = np.array(waveform.samples, dtype=np.csingle)

            pulse = scale_factor * amp * np.exp(1.0j * self._phase) * amplitude
            if not drag == 0.0:
                amplitude_differential = NumericFunction().derivative(t, amplitude)
                if len(amplitude_differential) < len(pulse):
                    amplitude_differential = np.pad(
                        amplitude_differential,
                        (0, len(pulse) - len(amplitude_differential)),
                        "edge",
                    )
                pulse += (
                    drag
                    * 1.0j
                    * amp
                    * scale_factor
                    * np.exp(1.0j * self._phase)
                    * amplitude_differential
                )
        else:
            pulse = waveform.sample(t, phase_offset=self._phase).samples

        # pulse = evaluate_shape(instruction.waveform, t, self._phase)

        scale = self.pulse_channel.scale
        if instruction.ignore_channel_scale:
            scale = 1
        pulse *= scale
        pulse += self.physical_channel.iq_voltage_bias.bias

        if do_upconvert:
            t += centre - 0.5 * dt + self._duration * dt
            pulse = self._do_upconvert(pulse, t)

        self._buffer[self._duration : self._duration + samples] = pulse
        self._duration += samples

    def _do_upconvert(self, buffer: list[float], time: list[float]):
        """A virtual NCO to upconvert the waveforms by a frequency that is the difference
        between the target frequency and the baseband frequency."""

        tslip = self.pulse_channel.phase_iq_offset
        imbalance = self.pulse_channel.imbalance
        freq = self._frequency - self.physical_channel.baseband.frequency
        buffer *= np.exp(UPCONVERT_SIGN * 2.0j * np.pi * freq * time)
        if not tslip == 0.0:
            buffer_slip = buffer * np.exp(UPCONVERT_SIGN * 2.0j * np.pi * freq * tslip)
            buffer.imag = buffer_slip.imag
        if not imbalance == 1.0:
            buffer.real /= imbalance**0.5
            buffer.imag *= imbalance**0.5

        return buffer

    def process_phaseshift(self, phase: float):
        self._phase += phase

    def process_phaseset(self, phase: float):
        self._phase = phase

    def process_phasereset(self):
        self._phase = 0

    def process_reset(self):
        log.warning(
            "The WaveformV1Backend uses a `repetition_time` for resetting, so the reset "
            "instruction will be ignored."
        )

    def process_frequencyset(self, frequency: float):
        self._frequency = frequency

    def process_frequencyshift(self, frequency: float):
        self._frequency += frequency

    def process_delay(self, samples: int):
        self._duration += samples


PydWaveformBackend = WaveformBackend



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from qat.model.device import (
    AcquirePulseChannel,
    CrossResonanceCancellationPulseChannel,
    CrossResonancePulseChannel,
    DrivePulseChannel,
    FreqShiftPulseChannel,
    MeasurePulseChannel,
    PhysicalBaseband,
    Qubit,
    QubitPhysicalChannel,
    QubitPulseChannels,
    Resonator,
    ResonatorPhysicalChannel,
    ResonatorPulseChannels,
    SecondStatePulseChannel,
)
from qat.model.hardware_model import PhysicalHardwareModel
from qat.utils.pydantic import CalibratableUnitInterval, FrozenDict, FrozenSet, QubitId


class PhysicalHardwareModelBuilder:
    """
    A builder class that builds a physical hardware model based on the given connectivity.

    :param physical_connectivity: The connectivities of the physical qubits on the QPU
        (undirected graph).
    :param logical_connectivity: The connectivities (directed graph) of the qubits used for
        compilation, which can be a subgraph of `physical_connectivity`.
    :param logical_connectivity_quality: Quality of the connections between the qubits.
    """

    def __init__(
        self,
        physical_connectivity: dict[int, FrozenSet[int]],
        logical_connectivity: dict[int, FrozenSet[int]] | None = None,
        logical_connectivity_quality: FrozenDict[
            tuple[QubitId, QubitId], CalibratableUnitInterval
        ] = None,
    ):
        self._current_model = self._build_uncalibrated_hardware_model(
            physical_connectivity=physical_connectivity,
            logical_connectivity_quality=logical_connectivity_quality,
            logical_connectivity=logical_connectivity,
        )

    @property
    def model(self) -> PhysicalHardwareModel:
        return self._current_model

    def _build_uncalibrated_hardware_model(
        self,
        physical_connectivity: dict[int, set[int]],
        logical_connectivity: dict[int, set[int]] = None,
        logical_connectivity_quality: None = None,
    ):
        logical_connectivity = logical_connectivity or physical_connectivity

        unique_qubit_indices = set(physical_connectivity.keys()).union(
            *physical_connectivity.values()
        )

        unused_chan_indices = set(range(len(unique_qubit_indices) * 2))

        qubits = {}
        for qubit_id in unique_qubit_indices:
            qubit_connectivity = physical_connectivity.get(qubit_id, set())

            bb_q = self._build_uncalibrated_baseband()
            bb_r = self._build_uncalibrated_baseband()

            physical_channel_q = self._build_uncalibrated_physical_channel(
                baseband=bb_q, target_device=Qubit, name_index=unused_chan_indices.pop()
            )
            physical_channel_r = self._build_uncalibrated_physical_channel(
                baseband=bb_r, target_device=Resonator, name_index=unused_chan_indices.pop()
            )

            pulse_channels_q = self._build_uncalibrated_qubit_pulse_channels(
                qubit_connectivity=qubit_connectivity
            )
            pulse_channels_r = self._build_uncalibrated_resonator_pulse_channels()

            resonator = Resonator(
                physical_channel=physical_channel_r, pulse_channels=pulse_channels_r
            )
            qubit = Qubit(
                physical_channel=physical_channel_q,
                pulse_channels=pulse_channels_q,
                resonator=resonator,
            )

            qubits[qubit_id] = qubit

        return PhysicalHardwareModel(
            qubits=qubits,
            logical_connectivity=logical_connectivity,
            logical_connectivity_quality=logical_connectivity_quality,
            physical_connectivity=physical_connectivity,
        )

    def _build_uncalibrated_baseband(self):
        return PhysicalBaseband()

    def _build_uncalibrated_physical_channel(
        self,
        baseband: PhysicalBaseband,
        target_device: type[Qubit | Resonator],
        name_index: int,
    ):
        if target_device is Qubit:
            return QubitPhysicalChannel(baseband=baseband, name_index=name_index)
        elif target_device is Resonator:
            return ResonatorPhysicalChannel(baseband=baseband, name_index=name_index)
        else:
            raise ValueError(f"Unsupported target device type: {target_device}")

    def _build_uncalibrated_qubit_pulse_channels(self, qubit_connectivity: list[QubitId]):
        cross_resonance_channels = FrozenDict(
            {
                q_other: CrossResonancePulseChannel(auxiliary_qubit=q_other)
                for q_other in qubit_connectivity
            }
        )
        cross_resonance_cancellation_channels = FrozenDict(
            {
                q_other: CrossResonanceCancellationPulseChannel(auxiliary_qubit=q_other)
                for q_other in qubit_connectivity
            }
        )

        pulse_channels = QubitPulseChannels(
            drive=DrivePulseChannel(),
            freq_shift=FreqShiftPulseChannel(),
            second_state=SecondStatePulseChannel(),
            cross_resonance_channels=cross_resonance_channels,
            cross_resonance_cancellation_channels=cross_resonance_cancellation_channels,
        )
        return pulse_channels

    def _build_uncalibrated_resonator_pulse_channels(self):
        pulse_channels_r = ResonatorPulseChannels(
            measure=MeasurePulseChannel(), acquire=AcquirePulseChannel()
        )
        return pulse_channels_r

    def model_dump(self):
        return self.model.model_dump()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import abc
from collections import defaultdict
from typing import Generic

from qat.core.metrics_base import MetricsManager
from qat.core.pass_base import PassManager
from qat.core.result_base import ResultManager
from qat.executables import Executable, Program
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PulseChannel
from qat.purr.compiler.hardware_models import QuantumHardwareModel


class BaseBackend(Generic[Program], abc.ABC):
    """
    Base class for a backend that takes an intermediate representation (IR) :class:`QatIR`
    and lowers it to machine code that can be executed on a given target.
    """

    def __init__(self, model: None | QuantumHardwareModel):
        """
        :param model: The hardware model that holds calibrated information on the qubits on the QPU.
        """
        self.model = model

    @abc.abstractmethod
    def emit(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        **kwargs,
    ) -> Executable[Program]:
        """
        Converts an IR :class:`QatIR` to machine instructions of a given target
        architecture.

        How targets convert the IR is at their discretion but they mostly follow
        macro-expansion techniques where target instructions are selected
        for each instruction in the IR.
        """
        ...


class CustomBackend(BaseBackend[Program], Generic[Program], abc.ABC):
    """
    Backends may need to run pre-codegen passes on the IR :class:`QatIR` as emitted
    from the middle end. These passes are specified via a custom pipeline.
    """

    def __init__(self, model: QuantumHardwareModel, pipeline: PassManager = None):
        super().__init__(model=model)
        self.pipeline = pipeline or PassManager()


class AllocatingBackend(CustomBackend[Program], Generic[Program], abc.ABC):
    """
    A backend that's responsible for allocating FPGA card and sequencers AOT.
    """

    def __init__(self, model: QuantumHardwareModel, pipeline: PassManager = None):
        super().__init__(model=model, pipeline=pipeline)
        self.allocations: dict[int, dict[str, int]] = defaultdict(dict)

    def allocate(self, target: PulseChannel) -> tuple[int, int]:
        """
        For a given target, allocate an FPGA card and a sequencer.
        """

        slot_idx = target.physical_channel.slot_idx
        target2seq = self.allocations[slot_idx]
        if (seq_idx := target2seq.get(target.full_id(), None)) is None:
            total = set(target.physical_channel.config.sequencers.keys())
            available = total - set(target2seq.values())
            if not available:
                raise ValueError(f"""
                No more available sequencers for {target}
                Physical channel: {target.physical_channel}
                Total available: {total}
                Current allocations: {target2seq}
                """)
            seq_idx = next(iter(available))
            target2seq[target.full_id()] = seq_idx
        return seq_idx, slot_idx



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from pydantic import BaseModel

from qat.ir.instructions import Instruction


class QatIR(BaseModel):
    """Stores in the intermediate representation of compiled quantum programs using QAT's
    instruction set.

    Currently only contains the list as a basic block.
    """

    instructions: list[Instruction] = []



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from collections import defaultdict
from enum import Enum
from random import Random

from qat.model.builder import PhysicalHardwareModelBuilder
from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.loaders.base import BasePhysicalModelLoader


class LucyCouplingDirection(Enum):
    """Enum for the coupling direction between qubits on a Lucy model."""

    LEFT = "left"
    RIGHT = "right"
    RANDOM = "random"


class LucyCouplingQuality(Enum):
    """Describes the coupling quality settings for a Lucy model."""

    UNIFORM = "uniform"
    RANDOM = "random"


class LucyModelLoader(BasePhysicalModelLoader):
    """Factory for creating mock Lucy hardware models. That is, a model of QPUs that have
    qubits arranged on a ring.

    The direction of couplings can be specified to go in either direction, or have each
    coupling choosen randomly. The coupling quality can be set to uniform, or random, or
    custom qualities can be provided as a dictionary mapping pairs of qubits to their
    coupling quality. Furthermore, some properties of the pulse channels can be provided.

    :param qubit_count: The number of qubits in the model. Default is 8.
    :param coupling_direction: The direction of the coupling between qubits. Default is
        `LucyCouplingDirection.RIGHT`.
    :param coupling_qualities: The coupling quality settings for the model. Default is
        `LucyCouplingQuality.UNIFORM`. Can also be a dictionary mapping pairs of qubits
        to their coupling quality.
    :param qubit_baseband_frequency: The baseband frequency for the qubits. Default is
        5.5 GHz.
    :param resonator_baseband_frequency: The baseband frequency for the resonators. Default
        is 8.5 GHz.
    :param qubit_baseband_if_frequency: The IF frequency for the qubits. Default is 250 MHz.
    :param resonator_baseband_if_frequency: The IF frequency for the resonators. Default
        is 250 MHz.
    :param drive_frequency: The frequency of the drive pulse channels. Default is 5.5 GHz.
    :param measure_frequency: The frequency of the measure pulse channels. Default is
        8.5 GHz.
    :param cr_scale: The scale factor for the cross-resonance pulse channels. Default is
        50.0.
    :param crc_scale: The scale factor for the cross-resonance cancellation pulse channels.
        Default is 0.0.
    :param random_seed: An optional seed for the random number generator to ensure
        reproducibility.
    :param start_index: The starting index for the qubits; useful for testing systems with
        non-standard indexing conventions. Default is 0.
    """

    def __init__(
        self,
        qubit_count: int = 8,
        coupling_direction: LucyCouplingDirection = LucyCouplingDirection.RIGHT,
        coupling_qualities: LucyCouplingQuality
        | dict[tuple[int, int], float] = LucyCouplingQuality.UNIFORM,
        qubit_baseband_frequency: float = 5.5e9,
        resonator_baseband_frequency: float = 8.5e9,
        qubit_baseband_if_frequency: float = 250e6,
        resonator_baseband_if_frequency: float = 250e6,
        drive_frequency: float = 5.5e9,
        measure_frequency: float = 8.5e9,
        cr_scale: float = 50.0,
        crc_scale: float = 0.0,
        random_seed: int | None = None,
        start_index: int = 0,
    ):
        self.qubit_count = qubit_count
        self.coupling_direction = coupling_direction
        self.coupling_qualities = coupling_qualities

        self.qubit_baseband_frequency = qubit_baseband_frequency
        self.resonator_baseband_frequency = resonator_baseband_frequency
        # TODO: remove if frequency? (COMPILER-714)
        self.qubit_baseband_if_frequency = qubit_baseband_if_frequency
        self.resonator_baseband_if_frequency = resonator_baseband_if_frequency
        self.drive_frequency = drive_frequency
        self.measure_frequency = measure_frequency
        self.cr_scale = cr_scale
        self.crc_scale = crc_scale
        self.start_index = start_index

        self.random_seed = random_seed
        self._random = Random(random_seed) if random_seed is not None else Random()

    def load(self) -> PhysicalHardwareModel:
        self._reset_random()
        model = self._make_uncalibrated_model()
        model = self._populate_physical_channels(model)
        return self._populate_pulse_channels(model)

    def _reset_random(self):
        """Resets the random number generator with the provided seed."""
        if self.random_seed is not None:
            self._random = Random(self.random_seed)
        else:
            self._random = Random()

    def _generate_physical_connectivity(self) -> dict[int, set[int]]:
        """Physical connectivity is bidirectional."""

        coupling_map = dict()
        for i in range(self.qubit_count):
            left_neighbor = self.start_index + (i - 1) % self.qubit_count
            right_neighbor = self.start_index + (i + 1) % self.qubit_count
            coupling_map[i + self.start_index] = {left_neighbor, right_neighbor}
        return coupling_map

    def _generate_logical_connectivity(self) -> dict[int, set[int]]:
        """Creates a logical connectivity that is unidirectional, with the direction decided
        by the coupling direction."""

        connectivity: dict[int, set[int]] = defaultdict(set)
        for i in range(self.qubit_count):
            left_neighbor = self.start_index + i
            right_neighbor = self.start_index + (i + 1) % self.qubit_count

            if self.coupling_direction == LucyCouplingDirection.RANDOM:
                direction = self._random.choice(
                    [LucyCouplingDirection.LEFT, LucyCouplingDirection.RIGHT]
                )
            else:
                direction = self.coupling_direction

            if direction == LucyCouplingDirection.LEFT:
                connectivity[right_neighbor].add(left_neighbor)
            elif direction == LucyCouplingDirection.RIGHT:
                connectivity[left_neighbor].add(right_neighbor)
        return connectivity

    def _order_custom_coupling_qualities(
        self, connectivity: dict[int, set[int]], qualities: dict[tuple[int, int], float]
    ) -> dict[tuple[int, int], float]:
        """Orders the coupling qualities according to the connectivity of the model."""

        ordered_qualities: dict[tuple[int, int], float] = {}
        for qubit1, qubit2 in qualities.keys():
            if qubit1 in connectivity and qubit2 in connectivity[qubit1]:
                ordered_qualities[(qubit1, qubit2)] = qualities[(qubit1, qubit2)]
            elif qubit2 in connectivity and qubit1 in connectivity[qubit2]:
                ordered_qualities[(qubit2, qubit1)] = qualities[(qubit1, qubit2)]
            else:
                raise ValueError(
                    f"Coupling quality for ({qubit1}, {qubit2}) does not match the ring "
                    "connectivity in Lucy models."
                )
        return ordered_qualities

    def _generate_coupling_qualities(
        self, connectivity: dict[int, set[int]]
    ) -> dict[tuple[int, int], float]:
        """Generates coupling qualities for the model, given the coupling quality type."""

        qualities = self.coupling_qualities
        if not isinstance(qualities, LucyCouplingQuality):
            return self._order_custom_coupling_qualities(connectivity, qualities)

        if qualities == LucyCouplingQuality.UNIFORM:
            return {
                (qubit1, qubit2): 1.0
                for qubit1, connections in connectivity.items()
                for qubit2 in connections
            }
        elif qualities == LucyCouplingQuality.RANDOM:
            return {
                (qubit1, qubit2): self._random.random()
                for qubit1, connections in connectivity.items()
                for qubit2 in connections
            }

    def _build_model(self, physical_connectivity, logical_connectivity, coupling_qualities):
        """Uses the connectivity and coupling qualities to build the model. Intentionally
        uses some indirection here so this can be adapted to other builders."""

        return PhysicalHardwareModelBuilder(
            physical_connectivity=physical_connectivity,
            logical_connectivity=logical_connectivity,
            logical_connectivity_quality=coupling_qualities,
        ).model

    def _make_uncalibrated_model(self) -> PhysicalHardwareModel:
        """Builds an uncalibrated hardware model based on the Lucy lattice."""

        physical_connectivity = self._generate_physical_connectivity()
        logical_connectivity = self._generate_logical_connectivity()
        coupling_qualities = self._generate_coupling_qualities(logical_connectivity)
        return self._build_model(
            physical_connectivity, logical_connectivity, coupling_qualities
        )

    def _populate_physical_channels(
        self, model: PhysicalHardwareModel
    ) -> PhysicalHardwareModel:
        """Populates the model with baseband frequencies for qubits and resonators."""

        for qubit in model.qubits.values():
            baseband = qubit.physical_channel.baseband
            baseband.frequency = self.qubit_baseband_frequency
            # TODO: remove if frequency? (COMPILER-714)
            baseband.if_frequency = self.qubit_baseband_if_frequency

            baseband = qubit.resonator.physical_channel.baseband
            baseband.frequency = self.resonator_baseband_frequency
            # TODO: remove if frequency? (COMPILER-714)
            baseband.if_frequency = self.resonator_baseband_if_frequency
        return model

    def _populate_pulse_channels(
        self, model: PhysicalHardwareModel
    ) -> PhysicalHardwareModel:
        """Populates the model with pulse channels and their frequencies and sclae factors."""

        for qubit in model.qubits.values():
            qubit.drive_pulse_channel.frequency = self.drive_frequency
            qubit.second_state_pulse_channel.frequency = 0.0
            qubit.freq_shift_pulse_channel.frequency = 0.0
            for channel in qubit.cross_resonance_pulse_channels.values():
                channel.frequency = self.drive_frequency
                channel.scale = self.cr_scale

            for channel in qubit.cross_resonance_cancellation_pulse_channels.values():
                channel.frequency = self.drive_frequency
                channel.scale = self.crc_scale

            qubit.measure_pulse_channel.frequency = self.measure_frequency
            qubit.acquire_pulse_channel.frequency = self.measure_frequency

        return model



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023-2025 Oxford Quantum Circuits Ltd
import uuid
from enum import Enum
from importlib.metadata import version

from compiler_config.config import InlineResultsProcessing

from qat.ir.instruction_builder import InstructionBuilder
from qat.purr.utils.logging_utils import log_duration

qir_available = True
try:
    from pyqir import (
        Call,
        Constant,
        Context,
        FloatConstant,
        IntConstant,
        Module,
        extract_byte_string,
        is_entry_point,
    )

    if version("pyqir") < "0.12.0":
        # TODO: Drop PyTket < 0.12.0 compatibility, COMPILER-909
        from pyqir import qubit_id, result_id

        def _convert_args(
            builder: InstructionBuilder,
            args: list[Constant],
            expected_types: list["ArgumentType"],
        ) -> list:
            """Converts QIR arguments to appropriate types for the instruction builder."""

            for i, arg in enumerate(args):
                if isinstance(arg, (IntConstant, FloatConstant)):
                    args[i] = arg.value
                elif (qid := qubit_id(arg)) is not None:
                    # Checks to see if this is a qubit
                    args[i] = builder.get_logical_qubit(qid)
                elif (rid := result_id(arg)) is not None:
                    # Checks to see if this is a result
                    args[i] = str(rid)
                elif expected_types[i] == ArgumentType.Label:
                    if not arg.is_null and (byte_string := extract_byte_string(arg)):
                        args[i] = byte_string.decode("utf-8").rstrip("\00")
                    else:
                        args[i] = None

            return args

    else:
        from pyqir import ptr_id

        def _convert_args(
            builder: InstructionBuilder,
            args: list[Constant],
            expected_types: list["ArgumentType"],
        ) -> list:
            """Converts QIR arguments to appropriate types for the instruction builder,
            given expected argument types.
            """

            for i, (arg, expected_type) in enumerate(zip(args, expected_types)):
                if expected_type == ArgumentType.Qubit:
                    if (qid := ptr_id(arg)) is not None:
                        args[i] = builder.get_logical_qubit(qid)
                    else:
                        raise ValueError(f"Argument {i} is not a valid qubit.")
                elif expected_type == ArgumentType.Result:
                    if (rid := ptr_id(arg)) is not None:
                        args[i] = str(rid)
                    else:
                        raise ValueError(f"Argument {i} is not a valid result.")
                elif expected_type == ArgumentType.Number:
                    if isinstance(arg, (IntConstant, FloatConstant)):
                        args[i] = arg.value
                    else:
                        raise ValueError(f"Argument {i} is not a valid number.")
                elif expected_type == ArgumentType.Label:
                    if not arg.is_null and (byte_string := extract_byte_string(arg)):
                        args[i] = byte_string.decode("utf-8").rstrip("\00")
                    else:
                        args[i] = None

            return args

except ImportError:
    qir_available = False


class ArgumentType(Enum):
    Qubit = 1
    Result = 2
    Number = 3
    Label = 4


class QIRParser:
    """Used to parse QIR files and walk the AST to assemble a quantum program.

    The parser makes API calls to the instruction builder passed at instantiation to allow
    for dynamic building of programs. This object is only responsible for interpreting
    and walking the QIR program.
    """

    def __init__(
        self,
        results_format: InlineResultsProcessing | None = None,
    ):
        self._results_format = (
            results_format
            if results_format is not None
            else InlineResultsProcessing.Program
        )
        self._result_variables = []

    def parse(self, builder: InstructionBuilder, qir_file: str):
        if not qir_available:
            raise RuntimeError("QIR parser unavailable.")

        with log_duration("QIR parsing completed, took {} seconds."):
            # Load the file if required
            if isinstance(qir_file, str):
                if qir_file.endswith(".bc"):
                    with open(qir_file, "rb") as f:
                        qir_file = f.read()
                elif qir_file.endswith(".ll"):
                    with open(qir_file) as f:
                        qir_file = f.read()

            # Create a module by inferring the type
            if isinstance(qir_file, str):
                mod = Module.from_ir(Context(), qir_file)
            elif isinstance(qir_file, bytes):
                mod = Module.from_bitcode(Context(), qir_file)
            else:
                raise ValueError(f"Expected type str | bytes, got {type(mod)} instead.")
            self._process_instructions(builder, mod)
            return builder

    @staticmethod
    def _throw_on_invalid_args(intrinsic_name: str, actual_args: int, expected_args: int):
        if actual_args != expected_args:
            raise ValueError(
                f"{intrinsic_name} has {actual_args} arguments, expected {expected_args}."
            )

    def _process_instruction(self, builder: InstructionBuilder, instruction):
        """Dispatches the correct calls to the builder given the type of instruction.

        Considered using a state dispatch approach, but the details get a little tricky
        and might add confusion. The benefit would be being able to write the instrinsic
        name next to explicit arguments and builder calls.
        """
        if not isinstance(instruction, Call):
            return

        intrinsic_name = instruction.callee.name
        args = instruction.args
        num_args = len(args)

        match intrinsic_name:
            case "__quantum__qis__ccx__body" | "__quantum____qis__ccnot__body":
                self._throw_on_invalid_args(intrinsic_name, num_args, 3)
                args = _convert_args(
                    builder,
                    args,
                    [ArgumentType.Qubit, ArgumentType.Qubit, ArgumentType.Qubit],
                )
                builder.ccnot(*args)
            case "__quantum__qis__cnot__body" | "__quantum__qis__cx__body":
                self._throw_on_invalid_args(intrinsic_name, num_args, 2)
                args = _convert_args(
                    builder, args, [ArgumentType.Qubit, ArgumentType.Qubit]
                )
                builder.cnot(*args)
            case "__quantum__qis__cz__body":
                self._throw_on_invalid_args(intrinsic_name, num_args, 2)
                args = _convert_args(
                    builder, args, [ArgumentType.Qubit, ArgumentType.Qubit]
                )
                builder.cZ(*args)
            case "__quantum__qis__h__body":
                self._throw_on_invalid_args(intrinsic_name, num_args, 1)
                args = _convert_args(builder, args, [ArgumentType.Qubit])
                builder.had(*args)
            case "__quantum__qis__mz__body":
                self._throw_on_invalid_args(intrinsic_name, num_args, 2)
                args = _convert_args(
                    builder, args, [ArgumentType.Qubit, ArgumentType.Result]
                )
                builder.measure_single_shot_z(args[0], output_variable=args[1])
                builder.results_processing(args[1], self._results_format)
            case "__quantum__qis__reset__body":
                self._throw_on_invalid_args(intrinsic_name, num_args, 1)
                args = _convert_args(builder, args, [ArgumentType.Qubit])
                builder.reset(*args)
            case "__quantum__qis__rx__body":
                self._throw_on_invalid_args(intrinsic_name, num_args, 2)
                args = _convert_args(
                    builder, args, [ArgumentType.Number, ArgumentType.Qubit]
                )
                builder.X(args[1], args[0])
            case "__quantum__qis__ry__body":
                self._throw_on_invalid_args(intrinsic_name, num_args, 2)
                args = _convert_args(
                    builder, args, [ArgumentType.Number, ArgumentType.Qubit]
                )
                builder.Y(args[1], args[0])
            case "__quantum__qis__rz__body":
                self._throw_on_invalid_args(intrinsic_name, num_args, 2)
                args = _convert_args(
                    builder, args, [ArgumentType.Number, ArgumentType.Qubit]
                )
                builder.Z(args[1], args[0])
            case "__quantum__qis__s__body":
                self._throw_on_invalid_args(intrinsic_name, num_args, 1)
                args = _convert_args(builder, args, [ArgumentType.Qubit])
                builder.S(*args)
            case "__quantum__qis__s_adj":
                self._throw_on_invalid_args(intrinsic_name, num_args, 1)
                args = _convert_args(builder, args, [ArgumentType.Qubit])
                builder.Sdg(*args)
            case "__quantum__qis__t__body":
                self._throw_on_invalid_args(intrinsic_name, num_args, 1)
                args = _convert_args(builder, args, [ArgumentType.Qubit])
                builder.T(*args)
            case "__quantum__qis__t__adj":
                self._throw_on_invalid_args(intrinsic_name, num_args, 1)
                args = _convert_args(builder, args, [ArgumentType.Qubit])
                builder.Tdg(*args)
            case "__quantum__qis__x__body":
                self._throw_on_invalid_args(intrinsic_name, num_args, 1)
                args = _convert_args(builder, args, [ArgumentType.Qubit])
                builder.X(*args)
            case "__quantum__qis__y__body":
                self._throw_on_invalid_args(intrinsic_name, num_args, 1)
                args = _convert_args(builder, args, [ArgumentType.Qubit])
                builder.Y(*args)
            case "__quantum__qis__z__body":
                self._throw_on_invalid_args(intrinsic_name, num_args, 1)
                args = _convert_args(builder, args, [ArgumentType.Qubit])
                builder.Z(*args)
            case "__quantum__rt__result_record_output":
                self._throw_on_invalid_args(intrinsic_name, num_args, 2)

                # not exactly clear if there is always a way to identify the type without
                # the instrinsic for context, so handle this here for now
                args = _convert_args(
                    builder,
                    args,
                    [ArgumentType.Result, ArgumentType.Label],
                )
                res = args[0]
                label = args[1]
                if label is None:
                    label = "generated_name_" + str(uuid.uuid4())
                self._result_variables.append((str(res), label))

            # currently do nothing; here for documentation purposes
            case "__quantum__rt__initialize":
                pass
            case "__quantum__rt__tuple_record_output":
                pass
            case "__quantum__rt__array_record_output":
                pass
            case _:
                pass

    def _process_instructions(self, builder: InstructionBuilder, module: Module):
        """Processes all instructions in the entry point function of the module."""

        entry_point = next(
            iter(filter(lambda fnc: is_entry_point(fnc), module.functions)), None
        )
        if entry_point is None:
            raise ValueError("Entry point unable to be found in QIR file.")

        instructions = [inst for bb in entry_point.basic_blocks for inst in bb.instructions]
        for inst in instructions:
            self._process_instruction(builder, inst)

        if any(self._result_variables):
            result_name = "_".join([val[1] for val in self._result_variables])

            builder.assign(result_name, [val[0] for val in self._result_variables])
            builder.returns(result_name)
        else:
            builder.returns()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.runtime.base import BaseRuntime as BaseRuntime
from qat.runtime.legacy import LegacyRuntime as LegacyRuntime
from qat.runtime.simple import SimpleRuntime as SimpleRuntime

DefaultRuntime = SimpleRuntime



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from pydantic import NonNegativeInt

from qat.ir.gates.base import QubitInstruction
from qat.ir.measure import AcquireMode


class Measure(QubitInstruction):
    """Instructs a measurement to be taken on the given qubit."""

    # TODO: decide if measure should have a "return type" (COMPILER-287)
    qubit: NonNegativeInt
    clbit: NonNegativeInt
    mode: AcquireMode = AcquireMode.INTEGRATOR

    @property
    def qubits(self):
        return (self.qubit,)


class Reset(QubitInstruction):
    r"""Instructs a qubit to be reset to it's lowest energy (:math:`|0>`) state."""

    # TODO: decide if reset should have a "reset type" (COMPILER-288)
    qubit: NonNegativeInt

    @property
    def qubits(self):
        return (self.qubit,)


class Barrier(QubitInstruction):
    """A barrier is a software construct that serves two purposes:
    #. Instructs the compiler to not optimize across the barrier (e.g. squashing of gates)
    #. Instructs the scheduler to synchronize targetted qubits up until this point.
    """

    qubits: list[NonNegativeInt]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023-2025 Oxford Quantum Circuits Ltd
from __future__ import annotations

from typing import Type, Union

import numpy as np
from pydantic import Field, model_validator

from qat.ir.instructions import QuantumInstruction
from qat.utils.pydantic import (
    AllowExtraFieldsModel,
    ComplexNDArray,
    FloatNDArray,
    find_all_subclasses,
)
from qat.utils.waveform import (
    BlackmanFunction,
    ComplexFunction,
    Cos,
    DragGaussianFunction,
    ExtraSoftSquareFunction,
    GaussianFunction,
    GaussianSquareFunction,
    GaussianZeroEdgeFunction,
    RoundedSquareFunction,
    SechFunction,
    SetupHoldFunction,
    Sin,
    SofterGaussianFunction,
    SofterSquareFunction,
    SoftSquareFunction,
    SquareFunction,
)


class AbstractWaveform(AllowExtraFieldsModel):
    @classmethod
    def name(cls) -> str:
        return cls.__name__.replace("Waveform", "")


class Waveform(AbstractWaveform):
    """
    Stores the attributes which define a waveform.
    """

    width: float = Field(ge=0, default=0)
    amp: float | complex = 0.0
    phase: float | complex = 0.0

    # All of the following are shape specific, we should probably deal with this better.
    drag: float = 0.0
    rise: float = 0.0
    amp_setup: float = 0.0
    scale_factor: float = 1.0
    zero_at_edges: bool = False
    beta: float = 0.0
    frequency: float = 0.0
    internal_phase: float = 0.0
    std_dev: float = 0.0
    square_width: float = 0.0

    shape_function_type: Type[ComplexFunction] | None = None

    @property
    def duration(self):
        return self.width

    def __repr__(self):
        return f"{self.__class__.__name__}(width={self.width}, amp={self.amp}, phase={self.phase})"

    def sample(self, t: np.ndarray, phase_offset: float = 0.0):
        if self.shape_function_type is None:
            raise AttributeError(
                f"Waveform of type `{self.__class__.__name__}` cannot be evaluated, please provide a valid shape function type."
            )

        # Generate a shape function based on the given type. The shape function
        # will only use whatever member attributes from this class it needs.
        shape_function = self.shape_function_type(**self.model_dump())

        amplitude = shape_function(t)
        phase_offset = self.phase + phase_offset
        samples = self.scale_factor * self.amp * np.exp(1.0j * phase_offset) * amplitude

        if self.drag:
            amplitude_differential = shape_function.derivative(t, amplitude)
            if len(amplitude_differential) < len(samples):
                amplitude_differential = np.pad(
                    amplitude_differential,
                    (0, len(samples) - len(amplitude_differential)),
                    "edge",
                )
            samples += (
                self.drag
                * 1.0j
                * self.amp
                * self.scale_factor
                * np.exp(1.0j * phase_offset)
                * amplitude_differential
            )

        return SampledWaveform(samples=samples)

    def __hash__(self):
        return hash(
            (
                self.__class__.__name__,
                self.width,
                self.amp,
                self.phase,
                self.drag,
                self.rise,
                self.amp_setup,
                self.scale_factor,
                self.zero_at_edges,
                self.beta,
                self.frequency,
                self.internal_phase,
                self.std_dev,
                self.square_width,
            )
        )


class SampledWaveform(AbstractWaveform):
    """
    Provide a list of amplitudes to define a sampled waveform.
    """

    samples: ComplexNDArray | FloatNDArray
    sample_time: float | None = None  # Time between samples, in seconds

    @property
    def duration(self):
        if self.sample_time is None:
            # TODO: COMPILER-723 -- Do we want to raise an error here, or return NAN or None or 0?
            raise ValueError(
                "Cannot determine duration of SampledWaveform without sample_time being set."
            )
        return self.sample_time * len(self.samples)

    def __repr__(self):
        return "sampled waveform"

    def __eq__(self, other: SampledWaveform):
        return np.array_equal(self.samples, other.samples)


class SquareWaveform(Waveform):
    shape_function_type: Type[SquareFunction] = SquareFunction


class SoftSquareWaveform(Waveform):
    shape_function_type: Type[SoftSquareFunction] = SoftSquareFunction


class SofterSquareWaveform(Waveform):
    shape_function_type: Type[SofterSquareFunction] = SofterSquareFunction


class ExtraSoftSquareWaveform(Waveform):
    shape_function_type: Type[ExtraSoftSquareFunction] = ExtraSoftSquareFunction


class GaussianWaveform(Waveform):
    shape_function_type: Type[GaussianFunction] = GaussianFunction


class SofterGaussianWaveform(Waveform):
    shape_function_type: Type[SofterGaussianFunction] = SofterGaussianFunction


class BlackmanWaveform(Waveform):
    shape_function_type: Type[BlackmanFunction] = BlackmanFunction


class SetupHoldWaveform(Waveform):
    shape_function_type: Type[SetupHoldFunction] = SetupHoldFunction


class RoundedSquareWaveform(Waveform):
    shape_function_type: Type[RoundedSquareFunction] = RoundedSquareFunction


class GaussianSquareWaveform(Waveform):
    shape_function_type: Type[GaussianSquareFunction] = GaussianSquareFunction


class DragGaussianWaveform(Waveform):
    shape_function_type: Type[DragGaussianFunction] = DragGaussianFunction


class GaussianZeroEdgeWaveform(Waveform):
    shape_function_type: Type[GaussianZeroEdgeFunction] = GaussianZeroEdgeFunction


class CosWaveform(Waveform):
    shape_function_type: Type[Cos] = Cos


class SinWaveform(Waveform):
    shape_function_type: Type[Sin] = Sin


class SechWaveform(Waveform):
    shape_function_type: Type[SechFunction] = SechFunction


waveform_classes = tuple(find_all_subclasses(Waveform) + [SampledWaveform])


class Pulse(QuantumInstruction):
    """
    Instructs a pulse channel to send a waveform. The intention of the waveform
    (e.g. a drive or measure pulse) can be specified using the type.
    """

    ignore_channel_scale: bool = False
    waveform: Union[waveform_classes]

    @model_validator(mode="before")
    def validate_duration(cls, data):
        # TODO: Review with COMPILER-723
        if isinstance(data, dict) and isinstance(data.get("waveform"), Waveform):
            data["duration"] = data["waveform"].duration
        return data

    def __repr__(self):
        return f"{self.__class__.__name__} on targets {set(self.targets)} with {self.waveform}."

    def update_duration(self, duration: float, sample_time: float | None = None):
        if isinstance(self.waveform, Waveform):
            self.duration = duration
            self.waveform.width = duration
        elif isinstance(self.waveform, SampledWaveform):
            if sample_time is None:
                sample_time = self.waveform.sample_time
            else:
                self.waveform.sample_time = sample_time

            current_duration = self.waveform.duration
            if current_duration > duration and not np.isclose(current_duration, duration):
                raise NotImplementedError(
                    "Cannot update the duration of a SampledWaveform to a smaller value. "
                    f"{current_duration} > {duration}\n"
                    "This would require removing samples, which is not supported."
                )
            # If the new duration is larger, we can pad the samples with zeros.
            padding = int(np.round((duration - current_duration) / sample_time, 0))
            self.waveform.samples = np.pad(
                self.waveform.samples,
                (0, padding),
                mode="constant",
                constant_values=0,
            )
            self.duration = self.waveform.duration
        else:
            raise ValueError(
                f"{type(self.waveform)} does not support updating duration. "
                "Can only apply with a Waveform or SampledWaveform."
            )

    @property
    def target(self):
        return next(iter(self.targets))

    @property
    def pulse_channel(self):
        return self.target



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from abc import ABC

import numpy as np

from qat.ir.gates.base import GateBase


class StateTensor(ABC):
    """A tensorial description of quantum-related objects, such as quantum states and
    unitary operators.

    Quantum many-body states with a discrete Hilbert space can be represented exactly using
    so state vector approaches, where the vector elements describe the probability amplitude
    for a particular state. The same concept can naturally be extended to described mixed
    states and quantum operations (unitary, non-unitary and quantum channels alike).

    This object is a minimal implementation of this type of reasoning. For now it is used
    purely for testing the higher parts of our compilation stack (hence its perhaps
    misplaced location in the `utils` package), such as gate-level optimisations. Its
    capabilities will be extended as required, and it could potentially be used to implement
    an in-house gate-level simulator later down the line...
    """

    pass


class StateVector(StateTensor):
    r"""Represents the wavefunction of a discrete quantum many-body state.

    The wavefunction can be represented as a :math:`d^{N}` vector for a system with
    :math:`N` constituents each with local dimension :math:`d^{N}`, which is :math:`d=2` for
    qubits. However, a more convenient (and entirely equivalent) way to represent this is
    actually as a :math:`(d, d, \dots, d, d)`-tensor (with total rank :math:`N`). This
    allows to do calculations using local operations (such as a 1Q or 2Q gate, or a
    measurement) without calculating the operation for the full Hilbert space, reducing the
    computational complexity of the operation. Note that while the cost is still exponential
    in :math:`N`, it can be done with a smaller exponent, allowing for more performant
    numerics.

    This object currently just represents the state an has implementations for basic
    application of quantum gates: this will be extended as needed.
    """

    def __init__(self, num_qubits: int):
        """Initiate the state with all qubits in their lowest-energy state."""

        self.num_qubits = num_qubits
        self.tensor = np.zeros([2] * num_qubits, dtype=np.complex128)
        self.tensor[(0,) * num_qubits] = 1.0

    def apply_gate(self, gate: GateBase) -> "StateVector":
        """Apply a gate to the quantum state."""
        self.tensor = self._apply_op(gate.matrix, gate.qubits)
        return self

    def _apply_op(self, matrix: np.ndarray, qubits: tuple[int, ...]):
        """Applies a matrix to the tensor, permuting the dimensions back into the correct
        order. Has computational cost O(2^(N+k)) where k is the number of qubits the matrix
        acts on."""

        qubits = (qubits,) if not isinstance(qubits, (tuple, list)) else qubits
        matrix = np.reshape(matrix, (2,) * (2 * len(qubits)))
        matrix_contraction_dims = tuple(len(qubits) + i for i in range(len(qubits)))
        tensor = np.tensordot(self.tensor, matrix, (qubits, matrix_contraction_dims))
        dim_order = tuple(set(range(self.num_qubits)) ^ set(qubits)) + qubits
        permute_order = sorted(range(len(dim_order)), key=lambda k: dim_order[k])
        return np.permute_dims(tensor, permute_order)


class StateOperator(StateTensor):
    r"""Represents a unitary operator for discrete quantum many-body quantum systems.

    A generalisation of :class:`StateVector` to matrix-like objects, which has rank
    :math:`2N`. Is initialized to be an identity matrix, which can be "evolved" by gates
    to represent a quantum circuit.
    """

    def __init__(self, num_qubits: int):
        """Initiate the the operator as the identity matrix."""

        self.num_qubits = num_qubits
        self.tensor = np.reshape(
            np.eye(2**num_qubits), tuple(2 for _ in range(2 * num_qubits))
        )

    def apply_gate(self, gate: GateBase) -> "StateOperator":
        """Apply a gate to the operator."""
        self.tensor = self._apply_op(gate.matrix, gate.qubits)
        return self

    def _apply_op(self, matrix: np.ndarray, qubits: tuple[int, ...]):
        """Applies a matrix to the tensor, permuting the dimensions back into the correct
        order. Has computational cost O(2^(2N+k)) where k is the number of qubits the matrix
        acts on."""

        qubits = (qubits,) if not isinstance(qubits, (tuple, list)) else qubits
        matrix = np.reshape(matrix, (2,) * (2 * len(qubits)))
        matrix_contraction_dims = tuple(len(qubits) + i for i in range(len(qubits)))
        tensor = np.tensordot(matrix, self.tensor, (matrix_contraction_dims, qubits))
        dim_order = (
            qubits
            + tuple(set(range(self.num_qubits)) ^ set(qubits))
            + tuple(self.num_qubits + i for i in range(self.num_qubits))
        )
        permute_order = sorted(range(len(dim_order)), key=lambda k: dim_order[k])
        return np.permute_dims(tensor, permute_order)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.model.loaders.purr.base import BaseLegacyModelLoader
from qat.purr.backends.realtime_chip_simulator import get_default_RTCS_hardware
from qat.purr.compiler.hardware_models import HardwareModel as LegacyHardwareModel


class RTCSModelLoader(BaseLegacyModelLoader):
    def __init__(self, rotating_frame: bool = True):
        self.rotating_frame = rotating_frame

    def load(self) -> LegacyHardwareModel:
        return get_default_RTCS_hardware(rotating_frame=self.rotating_frame)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import numpy as np

from qat.backend.waveform import WaveformProgram
from qat.engines.native import NativeEngine
from qat.purr.compiler.instructions import AcquireMode
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class EchoEngine(NativeEngine[WaveformProgram]):
    """The :class:`EchoEngine` is a minimal execution engine primarily used for testing the
    compilation pipeline.

    It is not connected to any target machine such as live hardware or a simulator, and just
    simply "echos" back the buffers. It only accepts :class:`WaveformProgram` programs.

    It checks the shape of the results generated match the expected shapes from codegen,
    to test consistency across the compilation and execution pipeline.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def execute(self, program: WaveformProgram, **kwargs) -> dict[str, np.ndarray]:
        """
        Execute a :class:`WaveformProgram`.

        :param program: The compiled program.
        :returns: The execution results.
        """
        results = {}
        acquire_shapes = program.acquire_shapes
        for channel_data in program.channel_data.values():
            buffer = np.asarray(channel_data.buffer)
            for acquire in channel_data.acquires:
                results[acquire.output_variable] = process_readout(
                    buffer[acquire.position : acquire.position + acquire.length],
                    program.shots,
                    acquire.mode,
                )

                if (
                    results[acquire.output_variable].shape
                    != acquire_shapes[acquire.output_variable]
                ):
                    raise ValueError(
                        f"Acquired data shape {results[acquire.output_variable].shape} "
                        f"does not match expected shape {acquire.shape} "
                        f"for variable {acquire.output_variable}."
                    )

        return results


def process_readout(readout: np.ndarray, shots: int, mode: AcquireMode):
    """
    Processes a single readout into the expected format for a given acquire mode.

    For :attr:`AcquireMode.RAW`, this means repeating the readout for a given number of
    shots. The :attr:`AcquireMode.INTEGRATOR` emulates the averaging on hardware by
    taking the mean; note that this does not currently include down scaling. Finally, the
    :attr:`AcquireMode.SCOPE` simply returns the readout back in its current form.
    """
    match mode:
        case AcquireMode.RAW:
            return np.tile(readout, shots).reshape((shots, -1))
        case AcquireMode.INTEGRATOR:
            return np.tile(np.mean(readout, axis=0), shots)
        case AcquireMode.SCOPE:
            return readout
    raise NotImplementedError(f"Acquire mode {mode} not currently supported.")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.model.loaders.purr.base import BaseLegacyModelLoader
from qat.purr.backends.live import build_lucy_hardware
from qat.purr.compiler.hardware_models import QuantumHardwareModel


class LucyModelLoader(BaseLegacyModelLoader):
    def load(self) -> QuantumHardwareModel:
        return build_lucy_hardware(QuantumHardwareModel())
